
# built in
import pandas as pd
import matplotlib.pyplot as plt
from enum import Enum
from typing import Optional

# qis
import qis.utils.dates as da
import qis.utils.np_ops as npn
import qis.perfstats.returns as ret
import qis.plots.time_series as pts
import qis.plots.utils as put
import qis.plots.heatmap as phe
import qis.models.linear.corr_cov_matrix as ccm
import qis.models.linear.ewm as ewm
from qis.models.linear.corr_cov_matrix import CorrMatrixOutput
from qis.models.linear.ewm import InitType
from qis.perfstats.config import PerfParams

from qis.perfstats.regime_classifier import add_bnb_regime_shadows, BenchmarkReturnsQuantileRegimeSpecs


def plot_corr_table(prices: pd.DataFrame,
                    var_format: str = '{:.0%}',
                    freq: Optional[str] = None,
                    cmap: str = 'PiYG',
                    ax: plt.Subplot = None,
                    **kwargs
                    ) -> plt.Figure:

    returns = ret.to_returns(prices=prices, is_log_returns=True, freq=freq)
    corr = ccm.compute_masked_covar_corr(returns=returns, is_covar=False)
    fig = phe.plot_heatmap(df=corr,
                           var_format=var_format,
                           cmap=cmap,
                           ax=ax,
                           **kwargs)
    return fig


def plot_ewm_corr_table(prices: pd.DataFrame,
                        ewm_lambda: float = 0.94,
                        var_format: str = '{:.0%}',
                        init_type: InitType = InitType.ZERO,
                        freq: Optional[str] = None,
                        cmap: str = 'PiYG',
                        ax: plt.Subplot = None,
                        is_average: bool = True,
                        **kwargs
                        ) -> plt.Figure:

    returns = ret.to_returns(prices=prices, is_log_returns=True, freq=freq)

    init_value = ewm.set_init_dim2(data=returns.to_numpy(), init_type=init_type)
    corr = ewm.compute_ewm_covar_tensor(a=returns.to_numpy(),
                                          ewm_lambda=ewm_lambda,
                                          covar0=init_value,
                                          is_corr=True)
    if is_average:
        ar = npn.tensor_mean(corr)
    else:
        ar = corr[-1]

    df = pd.DataFrame(ar, index=prices.columns, columns=prices.columns)
    fig = phe.plot_heatmap(df=df,
                           var_format=var_format,
                           cmap=cmap,
                           ax=ax,
                           **kwargs)
    return fig


def plot_corr_matrix_time_series(prices: pd.DataFrame,
                                 corr_matrix_output: CorrMatrixOutput = CorrMatrixOutput.FULL,
                                 time_period: da.TimePeriod = None,
                                 freq: str = 'B',
                                 ewm_lambda: float = 0.97,
                                 init_type: InitType = InitType.X0,
                                 var_format: str = '{:.0%}',
                                 legend_line_type: pts.LegendLineType = pts.LegendLineType.AVG_LAST,
                                 trend_line: put.TrendLine = put.TrendLine.AVERAGE,
                                 regime_benchmark_str: str = None,
                                 regime_params: BenchmarkReturnsQuantileRegimeSpecs = BenchmarkReturnsQuantileRegimeSpecs(),
                                 perf_params: PerfParams = None,
                                 ax: plt.Subplot = None,
                                 **kwargs
                                 ) -> None:

    """
    plot time series
    """
    returns = ret.to_returns(prices=prices, is_log_returns=False, freq=freq)
    corr_pandas = ccm.compute_ewm_corr_df(returns=returns,
                                          corr_matrix_output=corr_matrix_output,
                                          ewm_lambda=ewm_lambda,
                                          init_type=init_type)
    if time_period is not None:
        corr_pandas = time_period.locate(corr_pandas)

    pts.plot_time_series(df=corr_pandas,
                         legend_line_type=legend_line_type,
                         trend_line=trend_line,
                         var_format=var_format,
                         ax=ax,
                         **kwargs)
    if regime_benchmark_str is not None:
        if regime_benchmark_str in prices.columns:
            pivot_prices = prices[regime_benchmark_str].reindex(index=corr_pandas.index, method='ffill')
        else:
            raise KeyError(f"{regime_benchmark_str} not in {prices.columns}")

        if regime_benchmark_str is not None and regime_params is not None:
            add_bnb_regime_shadows(ax=ax,
                                   pivot_prices=pivot_prices,
                                   benchmark=regime_benchmark_str,
                                   regime_params=regime_params,
                                   perf_params=perf_params)


class UnitTests(Enum):
    CORR_TABLE = 1
    CORR_MATRIX = 2
    EWMA_CORR = 3


def run_unit_test(unit_test: UnitTests):

    from qis.data.yf_data import load_etf_data
    prices = load_etf_data().dropna()

    if unit_test == UnitTests.CORR_TABLE:
        plot_corr_table(prices=prices)

    elif unit_test == UnitTests.CORR_MATRIX:
        fig, ax = plt.subplots(1, 1, figsize=(10, 10), constrained_layout=True)
        plot_corr_matrix_time_series(prices=prices, regime_benchmark_str='SPY', ax=ax)

    elif unit_test == UnitTests.EWMA_CORR:
        plot_ewm_corr_table(prices=prices.iloc[:, :5])

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.CORR_MATRIX

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)


"""
plot histogram 2d
"""
# built in
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from scipy import stats
from typing import Union, List, Optional, Tuple
from enum import Enum

# qis
import qis.plots.utils as put


def plot_hist_plot2d(df: pd.DataFrame,
                     title: str = None,
                     a_min: float = None,
                     a_max: float = None,
                     xvar_format: str = '{:.1f}',
                     yvar_format: str = '{:.1f}',
                     add_corr_legend: bool = True,
                     legend_loc: Optional[str] = 'upper left',
                     color: str = 'navy',
                     fontsize: int = 10,
                     ax: plt.Subplot = None,
                     **kwargs
                     ) -> plt.Figure:

    if len(df.columns) != 2:
        raise ValueError(f"should be 2 columns")

    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
    else:
        fig = None

    if a_min is not None or a_max is not None:
        df = np.clip(df, a_min=a_min, a_max=a_max)

    sns.histplot(data=df,
                 x=df.columns[0],
                 y=df.columns[1],
                 bins=100,
                 cbar=False,
                 stat='probability',
                 cbar_kws=dict(shrink=.75),
                 ax=ax)

    put.set_ax_ticks_format(ax=ax, xvar_format=xvar_format, yvar_format=yvar_format, **kwargs)

    if add_corr_legend:
        rho, pval = stats.spearmanr(df.to_numpy(), nan_policy='omit', axis=0)  # column is variable
        label = f"Rank corr={rho:0.2f}, p-val={pval:0.2f}"
        lines = [(label, {'color': color})]

        put.set_legend(ax=ax,
                       legend_loc=legend_loc,
                       fontsize=fontsize,
                       lines=lines,
                       **kwargs)

    put.align_xy_limits(ax=ax)

    if title is not None:
        ax.set_title(title, fontsize=fontsize, **kwargs)

    return fig


class UnitTests(Enum):
    TEST = 1


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.TEST:
        np.random.seed(1)
        n_instruments = 1000
        exposures_nm = np.random.normal(0.0, 1.0, size=(n_instruments, 2))
        data = pd.DataFrame(data=exposures_nm, columns=[f"id{n+1}" for n in range(2)])

        fig, ax = plt.subplots(1, 1, figsize=(3.9, 3.4), tight_layout=True)
        global_kwargs = dict(fontsize=6, linewidth=0.5, weight='normal', first_color_fixed=True)
        plot_hist_plot2d(df=data, ax=ax, **global_kwargs)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.TEST

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




import functools
import timeit
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from typing import Union
from enum import Enum

from qis.data.ust_rates import load_ust_rates
import qis.models.linear.ewm as ewm
import qis.models.linear.auto_corr as ac
import qis.plots.utils as put
import qis.plots.time_series as pts


def plot_ust_rates_data():
    df = load_ust_rates()

    with sns.axes_style("darkgrid"):
        fig, ax = plt.subplots(1, 1, figsize=(10, 10), tight_layout=True)
        pts.plot_time_series(df=df,
                             legend_line_type=pts.LegendLineType.FIRST_AVG_LAST,
                             var_format='{:,.2f}',
                             ax=ax)

        fig, ax = plt.subplots(1, 1, figsize=(10, 10), tight_layout=True)
        pts.plot_time_series(df=df[['3m', '10y']],
                             legend_line_type=pts.LegendLineType.FIRST_AVG_LAST,
                             var_format='{:,.2f}',
                             ax=ax)


class UnitTests(Enum):
    PLOT_RATES = 1


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.PLOT_RATES:
        plot_ust_rates_data()

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PLOT_RATES

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



# built in
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from typing import Tuple, List, Dict
from enum import Enum

# qis
import qis.utils.dates as da
import qis.perfstats.returns as ret
import qis.plots.time_series as pts

# models
import qis.models.linear.ewm as ewm
import qis.models.stats.gaussian_mixture as gm
from qis.models.linear.corr_cov_matrix import compute_masked_covar_corr, matrix_regularization

# portfolio
from qis.perfstats import returns as ret
from qis.portfolio.optimization.qp_solvers import PortfolioObjective, max_portfolio_sharpe_qp
import qis.portfolio.backtester as bp


import qis.portfolio.optimization.rolling_portfolios as rlp
from qis.portfolio.portfolio_data import PortfolioData
from qis.utils import dates as da


from qis.data.yf_data import load_etf_data




class UnitTests(Enum):
    MAX_DIVERSIFICATION = 1


def run_unit_test(unit_test: UnitTests):

    # data
    prices = load_etf_data()


    import qis.plots.stackplot as pst

    kwargs = dict(is_add_mean_levels=True,
                  is_yaxis_limit_01=True,
                  baseline='zero',
                  bbox_to_anchor=(0.4, 1.1),
                  legend_line_type=pst.LegendLineType.AVG_STD_LAST,
                  ncol=len(prices.columns)//3,
                  var_format='{:.0%}')

    if unit_test == UnitTests.MAX_DIVERSIFICATION:
        weights = rlp.solve_optimal_weights_ewm_covar(prices=prices,
                                                      portfolio_objective=PortfolioObjective.MAX_DIVERSIFICATION,
                                                      weight_mins=np.zeros(len(prices.columns)),
                                                      weight_maxs=np.ones(len(prices.columns)),
                                                      rebalancing_freq='Q',
                                                      ewm_lambda=0.97)

        portfolio_out = bp.backtest_model_portfolio(prices=prices,
                                                    weights=weights,
                                                    is_rebalanced_at_first_date=True,
                                                    ticker='MaxDiversification',
                                                    is_output_portfolio_data=True)


        portfolio_out.plot_nav()
    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.MAX_DIVERSIFICATION

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
provide generic for data appending set data savings
"""
import pandas as pd
from dataclasses import dataclass
from typing import Optional, Dict
import qis.file_utils as fu


@dataclass
class DfOutDict:
    df_out_dict: Dict[str, pd.DataFrame] = None
    last_data: Optional[pd.DataFrame] = None

    def __post_init__(self):
        self.df_out_dict = {}

    def __getitem__(self, key: str) -> pd.DataFrame:
        # for df_out_dict[key]
        return self.df_out_dict[key]

    def append(self, df: pd.DataFrame, name: str) -> None:
        if name not in self.df_out_dict.keys():
            self.df_out_dict[name] = df
        else:
            raise ValueError(f"{name} exist in {self.df_out_dict.keys()}")

    def set_last_df(self, data: pd.DataFrame) -> None:
        # way to first make a placement and then append it
        self.last_data = data

    def append_last_df(self, name: str) -> None:
        # way to first make a placement and then append it
        if self.last_data is None:
            raise ValueError(f"self.last_data is None")
        self.append(df=self.last_data, name=name)
        self.last_data = None

    def print(self):
        for key, df in self.df_out_dict.items():
            print(f"{key}")
            print(df)

    def save(self, file_name: str) -> None:
        if len(self.df_out_dict.keys()) > 0:
            file_path = fu.save_df_to_excel(data=self.df_out_dict, file_name=file_name)
            print(f"saved output data to excel:\n {file_path}")



# built in
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from enum import Enum

# qis
import qis.utils.dates as da
import qis.plots.derived.returns_heatmap as rhe
import qis.plots.derived.prices as pdp
import qis.perfstats.perf_table as rpt
from qis.utils.dates import TimePeriod
import qis.plots.table as ptb
import qis.plots.derived.regime_scatter as drc
from qis.plots.derived.perf_table import plot_ra_perf_table
from qis.perfstats.config import PerfParams

# data
from qis.data.yf_data import load_etf_data, fetch_prices
from qis.data.ust_rates import load_ust_3m_rate


def generate_performances(prices: pd.DataFrame,
                          regime_benchmark_str: str,
                          perf_params: PerfParams = None,
                          performance_label: pdp.PerformanceLabel = pdp.PerformanceLabel.WITH_DD,
                          ) -> None:

    kwargs = dict(digits_to_show=1, legend_alpha=0.75, performance_label=performance_label)

    fig, ax = plt.subplots(1, 1, figsize=(12, 4), tight_layout=True)
    plot_ra_perf_table(prices=prices,
                       perf_columns=rpt.EXTENDED_TABLE_COLUMNS,
                       perf_params=perf_params,
                       ax=ax)

    fig, ax = plt.subplots(1, 1, figsize=(6, ptb.calc_table_height(num_rows=len(prices.columns), scale=0.4)), tight_layout=True)
    rhe.plot_periodic_returns_table_from_prices(prices=prices,
                                                freq='A',
                                                ax=ax,
                                                title=f"Monthly Performance: {da.get_time_period_label(prices, date_separator='-')}",
                                                total_name='YTD',
                                                **{'square': False, 'x_rotation': 90})

    with sns.axes_style("darkgrid"):
        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
        pdp.plot_prices(prices=prices,
                        regime_benchmark_str=regime_benchmark_str,
                        perf_params=perf_params,
                        ax=ax,
                        **kwargs)

        fig, axs = plt.subplots(2, 1, figsize=(7, 7))
        pdp.plot_prices_with_dd(prices=prices,
                                regime_benchmark_str=regime_benchmark_str,
                                perf_params=perf_params,
                                axs=axs,
                                **kwargs)

        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
        drc.plot_scatter_regression(prices=prices,
                                    regime_benchmark_str=regime_benchmark_str,
                                    perf_params=perf_params,
                                    title='Regime Conditional Regression',
                                    ax=ax,
                                    **kwargs)


class UnitTests(Enum):
    ETF_DATA = 1


def run_unit_test(unit_test: UnitTests):

    prices = load_etf_data().dropna()
    ust_3m_rate = load_ust_3m_rate()

    perf_params = PerfParams(freq='W-WED', freq_reg='M', freq_drawdown='B', rates_data=ust_3m_rate)

    if unit_test == UnitTests.ETF_DATA:
        generate_performances(prices=prices,
                              perf_params=perf_params,
                              regime_benchmark_str='SPY')

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.ETF_DATA

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)





import functools
import timeit
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Union
from enum import Enum

import qis.models.linear.ewm as ewm
import qis.models.linear.auto_corr as ac
import qis.plots.utils as put
import qis.plots.time_series as pts


def get_test_data(n1: int = 1000, n2: int = 3, means: Union[float, np.ndarray] = 0.0) -> pd.DataFrame:
    np_data = np.random.normal(means, 1.0, (n1, n2))
    random_nans = np.random.binomial(size=(n1, n2), n=1, p=0.5)
    np_data = np.where(random_nans, np.nan, np_data)
    df = pd.DataFrame(data=np_data,
                      index=pd.date_range(start='1Jan2020', periods=n1, freq='D'),
                      columns=[f"x{m+1}" for m in range(n2)])
    return df


def time_comp_with_pandas_ewm(span: int = 31) -> None:

    big_df = get_test_data(n1=10000, n2=10000)
    np_data = big_df.to_numpy()
    init_value = np.zeros(np_data.shape[1])

    def pd_ewm():
        big_df.ewm(span=span, adjust=False).mean()

    def np_to_ewm():
        ewm.ewm_recursion(a=np_data, span=span, init_value=init_value)

    def pd_to_ewm():
        ewm.compute_ewm(data=big_df, span=span, init_value=init_value)

    n = 20
    print(f"pandas ewm: {timeit.Timer(functools.partial(pd_ewm)).timeit(n) / n:.3f}")
    print(f"np_to_ewm: {timeit.Timer(functools.partial(np_to_ewm)).timeit(n) / n:.3f}")
    print(f"pd_to_ewm: {timeit.Timer(functools.partial(pd_to_ewm)).timeit(n) / n:.3f}")


def vol_comp_with_pandas_ewm(span: int = 31) -> None:
    df = get_test_data(n1=10000, n2=1)
    vol_pd = df.ewm(span=span, adjust=False).std()
    vol_ewm1 = ewm.compute_ewm_vol(data=df, span=span, mean_adj_type=ewm.MeanAdjType.NONE)
    vol_ewm2 = ewm.compute_ewm_vol(data=df, span=span, mean_adj_type=ewm.MeanAdjType.EWMA)
    joint = pd.concat([vol_pd.iloc[:, 0].rename('Pandas std'),
                      vol_ewm1.iloc[:, 0].rename('Ewm no adj'),
                      vol_ewm2.iloc[:, 0].rename('Ewm with Ewm adj')],
                      axis=1)

    pts.plot_time_series(df=joint,
                         title='Vols comparision',
                         legend_loc='upper left',
                         trend_line=put.TrendLine.AVERAGE,
                         var_format='{:.2f}')


class UnitTests(Enum):
    TIME_COMP_WITH_PANDAS_EWM = 1
    VOL_COMP_WITH_PANDAS_EWM = 2
    EWMA_NP = 3
    EWMA_DF = 4
    EWMA_MEAN = 5
    EWMA_VOL = 6
    EWMA_BETA = 7
    EWMA_AUTO_CORR = 8
    EWMA_CORR_MATRIX = 9


def run_unit_test(unit_test: UnitTests):

    np.random.seed(1)

    if unit_test == UnitTests.TIME_COMP_WITH_PANDAS_EWM:
        time_comp_with_pandas_ewm()

    elif unit_test == UnitTests.VOL_COMP_WITH_PANDAS_EWM:
        vol_comp_with_pandas_ewm()

    else:

        data = get_test_data(n1=10000, n2=3, means=np.array([-2.0, -1.0, 0.0]))
        ewm_lambda = 0.94
        ewm_lambda3 = np.array([0.94, 0.50, 0.10])
        plot_data, title = None, None

        if unit_test == UnitTests.EWMA_NP:
            ewm1 = ewm.ewm_recursion(a=data.iloc[:, 0].to_numpy(dtype=np.double), ewm_lambda=ewm_lambda, init_value=0.0)
            print(ewm1)

            ewm2 = ewm.ewm_recursion(a=data.to_numpy(), ewm_lambda=ewm_lambda, init_value=np.zeros(len(data.columns)))
            print(ewm2)

            ewm3 = ewm.ewm_recursion(a=data.to_numpy(), ewm_lambda=ewm_lambda3, init_value=np.zeros(len(data.columns)))
            print(ewm3)

        elif unit_test == UnitTests.EWMA_DF:

            ewm_df1 = ewm.compute_ewm(data=data.iloc[:, 0], ewm_lambda=ewm_lambda, init_type=ewm.InitType.MEAN)
            print(ewm_df1)

            ewm_df1_np = ewm.compute_ewm(data=data.iloc[:, 0].to_numpy(), ewm_lambda=ewm_lambda, init_type=ewm.InitType.MEAN)
            print(ewm_df1_np)

            ewm_df2_np = ewm.compute_ewm(data=data.to_numpy(), ewm_lambda=ewm_lambda, init_type=ewm.InitType.MEAN)
            print(ewm_df2_np)

            ewm_df2 = ewm.compute_ewm(data=data, ewm_lambda=ewm_lambda, init_type=ewm.InitType.MEAN)
            print(ewm_df2)

            ewm_df3 = ewm.compute_ewm(data=data, ewm_lambda=ewm_lambda3, init_type=ewm.InitType.MEAN)
            print(ewm_df3)

            ewm_df3.columns = [x + ' ewm' for x in data.columns]
            plot_data = pd.concat([data, ewm_df3], axis=1)
            title = 'ewm'

        elif unit_test == UnitTests.EWMA_MEAN:

            datas = {'np1': data.iloc[:, 0].to_numpy(), 'np': data.to_numpy(), 'series': data.iloc[:, 0], 'df': data}

            for key, data_ in datas.items():
                print(key)
                ewm_mean = ewm.compute_roll_mean(data=data,
                                                 mean_adj_type=ewm.MeanAdjType.EWMA,
                                                 ewm_lambda=ewm_lambda)
                data_adjusted = data - ewm_mean
                ewm_data = ewm.compute_ewm(data=data_adjusted,
                                              ewm_lambda=ewm_lambda)

            ewm_mean.columns = [x + ' ewm_mean' for x in data.columns]
            ewm_data.columns = [x + '-ewm_mean' for x in data.columns]
            plot_data = pd.concat([data, ewm_mean, ewm_data], axis=1)
            title = 'ewm-mean'

        elif unit_test == UnitTests.EWMA_VOL:

            datas = {'np1': data.iloc[:, 0].to_numpy(), 'np': data.to_numpy(), 'series': data.iloc[:, 0], 'df': data}

            for key, data_ in datas.items():
                print(key)
                ewm_data = ewm.compute_ewm_vol(data=data_,
                                                  ewm_lambda=ewm_lambda,
                                                  mean_adj_type=ewm.MeanAdjType.EWMA,
                                                  init_type=ewm.InitType.MEAN)
            print(ewm_data)
            ewm_data.columns = [x + ' ewm vol' for x in data.columns]
            data.columns = [x + '^2' for x in data.columns]
            plot_data = pd.concat([np.power(data, 2), ewm_data], axis=1)
            title = 'ewm vol'

        elif unit_test == UnitTests.EWMA_BETA:

            cross_xy_types = [ewm.CrossXyType.COVAR, ewm.CrossXyType.BETA, ewm.CrossXyType.CORR]

            for cross_xy_type in cross_xy_types:
                print(cross_xy_type)
                ewm_data = ewm.compute_ewm_cross_xy(x_data=data[[data.columns[0]]],
                                                       y_data=data,
                                                       cross_xy_type=cross_xy_type,
                                                       ewm_lambda=ewm_lambda,
                                                       mean_adj_type=ewm.MeanAdjType.EWMA,
                                                       init_type=ewm.InitType.MEAN)
            ewm_data.columns = [x + ' ewm beta' for x in data.columns]
            plot_data = pd.concat([data, ewm_data], axis=1)
            title = 'ewm beta'

        elif unit_test == UnitTests.EWMA_AUTO_CORR:
            plot_data = ac.compute_dynamic_auto_corr(data=data,
                                                     ewm_lambda=ewm_lambda,
                                                     mean_adj_type=ewm.MeanAdjType.NONE,
                                                     aggregation_type='median',
                                                     is_normalize=True)
            title = 'ewm auto-corr'

        else:
            return

        if plot_data is not None:
            pts.plot_time_series(df=plot_data,
                                 title=title,
                                 legend_loc='upper left',
                                 trend_line=put.TrendLine.AVERAGE,
                                 var_format='{:.2f}')

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.EWMA_NP

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



from enum import Enum


class EnumMap(Enum):
    """
    abstract enum with a map function
    """
    @classmethod
    def map_to_value(cls, name):
        """
        given name return value
        """
        for k, v in cls.__members__.items():
            if k == name:
                return v
        raise ValueError(f"nit in enum {name}")


"""
implementation of dataclass with save method
"""
import pandas as pd
from dataclasses import dataclass
from typing import get_type_hints
import qis.file_utils as fu


@dataclass
class DataClassSave:
    pass

    def save_to_excel(self,
                      file_name: str,
                      is_transpose: bool = False,
                      indedate_format: str = None): #'%d%b%Y'
        data = {}
        for field_name, field_type in get_type_hints(self).items():
            if issubclass(field_type, pd.DataFrame):
                df = getattr(self, field_name)
                if df is not None:
                    if indedate_format is not None:
                        df.index = [x.strftime(indedate_format) if isinstance(x, pd.Timestamp) else x for x in df.index]
                    if is_transpose:
                        df = df.transpose()
                    data[field_name] = df
        fu.save_df_to_excel(data=data, file_name=file_name)



"""
scatter plot core
"""
# built in
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import scipy.stats as stats
from statsmodels import api as sm
from typing import Union, List, Tuple, Optional

# qis
import qis.plots.utils as put
from qis.utils.ols import get_ols_x, reg_model_params_to_str


def plot_scatter(df: pd.DataFrame,
                 x_column: str = None,
                 y_column: str = None,
                 hue: str = None,
                 xlabel: Union[str, bool, None] = True,
                 ylabel: Union[str, bool, None] = True,
                 title: Optional[str] = None,
                 use_regplot: bool = False,
                 annotation_labels: List[str] = None,
                 annotation_colors: List[str] = None,
                 annotation_color: Optional[str] = 'red',
                 add_universe_model_label: bool = True,
                 add_universe_model_prediction: bool = False,
                 add_universe_model_ci: bool = False,
                 add_hue_model_label: bool = False,  # add hue eqs for data with hue
                 ci: Optional[int] = None,
                 order: int = 1,  # regression order
                 fit_intercept: bool = True,
                 full_sample_order: Optional[int] = None,  # full sample order can be different
                 color0: str = 'darkblue',
                 colors: List[str] = None,
                 xvar_format: str = '{:.0%}',
                 yvar_format: str = '{:.0%}',
                 x_limits: Tuple[Optional[float], Optional[float]] = None,
                 y_limits: Tuple[Optional[float], Optional[float]] = None,
                 xticks: List[str] = None,
                 fontsize: int = 10,
                 linewidth: float = 1.5,
                 markersize: int = 4,
                 first_color_fixed: bool = False,
                 full_sample_label: str = 'Full sample: ',
                 is_add_45line: bool = False,
                 is_r2_only: bool = False,
                 legend_loc: str = 'upper left',
                 value_name: str = 'value_name',
                 ax: plt.Subplot = None,
                 **kwargs
                 ) -> plt.Figure:
    df = df.copy().dropna()

    if x_column is None:
        if len(df.columns) == 2:
            x_column = df.columns[0]
        else:
            raise ValueError(f"x_column is not defined for more than on columns")
    if y_column is None:
        if len(df.columns) == 2:  # x and y
            y_column = df.columns[1]
        else:  # melting to column value_name with hue = all columns ba t x
            hue = 'hue'
            add_hue_model_label = True
            y_column = value_name
            df = pd.melt(df, id_vars=[x_column], value_vars=df.columns.drop(x_column), var_name=hue,
                         value_name=value_name)

    if full_sample_order is None:
        full_sample_order = order

    if add_universe_model_label:
        hue_offset = 1
    else:
        hue_offset = 0

    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
    else:
        fig = None

    estimated_reg_models = {}

    if hue is not None:
        if colors is None:
            colors = put.get_n_colors(n=len(df[hue].unique()), first_color_fixed=first_color_fixed)

        if hue_offset == 1:
            palette = colors #[color0] +
        else:
            palette = colors

        hue_ids = df[hue].unique()
        for idx, hue_id in enumerate(hue_ids):
            # estimate model equation
            data_hue = df[df[hue] == hue_id].sort_values(by=x_column)
            x = data_hue[x_column].to_numpy(dtype=np.float)
            y = data_hue[y_column].to_numpy(dtype=np.float)
            x1 = get_ols_x(x=x, order=order, fit_intercept=fit_intercept)
            reg_model = sm.OLS(y, x1).fit()
            estimated_reg_models[hue_id] = reg_model

            # plot data points
            #try:
            sns.scatterplot(x=x_column, y=y_column, data=data_hue, color=palette[idx], s=markersize, ax=ax)
            #except:
            #    print(f"scatter plot problem with {data_hue}")

            if order > 0:  # plot prediction
                if use_regplot:
                    # not possible to control reg equation in regplot
                    sns.regplot(x=x_column, y=y_column, data=data_hue,
                                ci=ci,
                                order=order, truncate=True,
                                color=palette[idx],
                                scatter_kws={'s': markersize},
                                line_kws={'linewidth': linewidth},
                                ax=ax)
                else:
                    prediction = reg_model.predict(x1)
                    ax.plot(x, prediction, color=palette[idx], lw=linewidth, linestyle='-')

    else:
        if full_sample_order == 0:  # just scatter plot
            sns.scatterplot(x=x_column, y=y_column, data=df,
                            # ci=ci,
                            s=markersize, color=color0, ax=ax)
        else:  # regplot add scatter and ml lines even if order is == 0
            sns.regplot(x=x_column, y=y_column, data=df, ci=ci, order=full_sample_order, color=color0,
                        scatter_kws={'s': markersize},
                        line_kws={'linewidth': linewidth}, ax=ax)

    # add ml equations to labels
    legend_labels = []
    legend_colors = []
    if (add_universe_model_prediction or add_universe_model_label or add_universe_model_ci) and order > 0:
        xy = df[[x_column, y_column]].sort_values(by=x_column)
        x = xy[x_column].to_numpy(dtype=np.float)
        y = xy[y_column].to_numpy(dtype=np.float)
        x1 = get_ols_x(x=x, order=full_sample_order, fit_intercept=fit_intercept)
        reg_model = sm.OLS(y, x1).fit()

        if add_universe_model_prediction:
            prediction = reg_model.predict(x1)
            ax.plot(x, prediction, color=color0, lw=linewidth, linestyle='--')

        if add_universe_model_ci:
            y_model = reg_model.predict(x1)
            ci = calc_ci(x=x, y=y, y_model=y_model)
            # ax.fill_between(x, y + ci, y - ci, color="None", linestyle="--")
            ax.plot(x, y_model - ci, "--", color="0.5")
            ax.plot(x, y_model + ci, "--", color="0.5")

        if add_universe_model_label:
            text_str = f"{full_sample_label} " \
                       f"{reg_model_params_to_str(reg_model=reg_model, order=full_sample_order, is_r2_only=False, fit_intercept=fit_intercept, **kwargs)}"
            legend_labels.append(text_str)
            legend_colors.append(color0)

    # add colors for annotation labels
    df['color'] = color0
    if hue is not None and add_hue_model_label:
        hue_ids = df[hue].unique()
        for color, hue_id in zip(colors, hue_ids):
            df.loc[df[hue] == hue_id, 'color'] = 'red'  # ad color for hue
            if order > 0:
                reg_model = estimated_reg_models[hue_id]
                text_str = (f"{hue_id}: " 
                            f"{reg_model_params_to_str(reg_model=reg_model, order=order, is_r2_only=is_r2_only, fit_intercept=fit_intercept, **kwargs)}")
                legend_labels.append(text_str)
            else:
                legend_labels.append(hue_id)
            legend_colors.append(color)

    elif hue is not None and order == 0:
        legend_labels = df[hue].unique()
        legend_colors = colors

    # add labels
    if annotation_labels is not None:
        if annotation_colors is not None:
            colors = annotation_colors
        elif annotation_color is not None:
            colors = len(df.index) * [annotation_color]
        else:
            colors = df['color']
        for label, x, y, color in zip(annotation_labels, df[x_column], df[y_column], colors):
            ax.annotate(label,
                        xy=(x, y), xytext=(1, 1),
                        textcoords='offset points', ha='left', va='bottom',
                        color=color,
                        fontsize=fontsize)
            if label != '':
                ax.scatter(x=x, y=y, c=color, s=20)

    if is_add_45line:  # make equal:
        ymin, ymax = ax.get_ylim()
        xmin, xmax = ax.get_xlim()
        min = ymin if ymin < xmin else xmin
        max = ymax if ymax > xmax else xmax
        ax.set_xlim([min, max])
        ax.set_ylim([min, max])
        x = np.linspace(*ax.get_xlim())
        ax.plot(x, x, color='black', lw=1, linestyle='--')

    if x_limits is not None:
        put.set_x_limits(ax=ax, x_limits=x_limits)
    if y_limits is not None:
        put.set_y_limits(ax=ax, y_limits=y_limits)

    if xticks is not None:
        put.set_ax_tick_labels(ax=ax, x_rotation=0, xticks=df[x_column].to_numpy(), x_labels=xticks,
                               fontsize=fontsize)
    else:
        put.set_ax_tick_labels(ax=ax, x_rotation=0, fontsize=fontsize)

    if isinstance(xlabel, bool) and xlabel is True:
        xlabel = 'x = ' + x_column
    if isinstance(ylabel, bool) and ylabel is True:
        ylabel = 'y = ' + y_column
    put.set_ax_xy_labels(ax=ax, xlabel=xlabel, ylabel=ylabel, fontsize=fontsize, **kwargs)

    put.set_ax_ticks_format(ax=ax, xvar_format=xvar_format, yvar_format=yvar_format, fontsize=fontsize, **kwargs)

    put.set_legend(ax=ax,
                   labels=legend_labels,
                   colors=legend_colors,
                   legend_loc=legend_loc,
                   fontsize=fontsize,
                   **kwargs)

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize, **kwargs)

    put.set_spines(ax=ax, **kwargs)

    return fig


def calc_ci(x: np.ndarray, y: np.ndarray, y_model: np.ndarray) -> np.ndarray:
    n = x.shape[0]
    m = 2
    dof = n - m
    t = stats.t.ppf(0.95, dof)
    #x2 = np.linspace(np.min(x), np.max(x), 100)

    # Estimates of Error in Data/Model
    resid = y - y_model
    # chi2 = np.sum((resid / y_model) ** 2)  # chi-squared; estimates error in data
    # chi2_red = chi2 / dof  # reduced chi-squared; measures goodness of fit
    s_err = np.sqrt(np.sum(resid ** 2) / dof)  # standard deviation of the error

    ci = t * s_err * np.sqrt(1 / n + (x - np.mean(x)) ** 2 / np.sum((x - np.mean(x)) ** 2))
    return ci



"""
compute risk-adjusted performance tables
"""
# built in
import numpy as np
import pandas as pd
from scipy.stats import kurtosis, skew
from typing import Callable, Union, Dict, Tuple
from collections import OrderedDict
from enum import Enum

# qis
import qis.utils.dates as da
import qis.utils.ols as ols
import qis.perfstats.returns as ret
import qis.perfstats.drawdowns as spd
from qis.perfstats.config import PerfStat, PerfParams


STANDARD_TABLE_COLUMNS = (PerfStat.START_DATE,
                          PerfStat.END_DATE,
                          PerfStat.PA_RETURN,
                          PerfStat.VOL,
                          PerfStat.SHARPE,
                          PerfStat.MAX_DD,
                          PerfStat.MAX_DD_VOL,
                          PerfStat.SKEWNESS,
                          PerfStat.KURTOSIS)

LN_TABLE_COLUMNS = (PerfStat.START_DATE,
                    PerfStat.END_DATE,
                    PerfStat.TOTAL_RETURN,
                    PerfStat.PA_RETURN,
                    PerfStat.AN_LOG_RETURN,
                    PerfStat.VOL,
                    PerfStat.SHARPE,
                    PerfStat.SHARPE_LOG_AN,
                    PerfStat.MAX_DD,
                    PerfStat.MAX_DD_VOL,
                    PerfStat.SKEWNESS,
                    PerfStat.KURTOSIS)

LN_BENCHMARK_TABLE_COLUMNS = (PerfStat.START_DATE,
                              PerfStat.END_DATE,
                              PerfStat.TOTAL_RETURN,
                              PerfStat.PA_RETURN,
                              PerfStat.AN_LOG_RETURN,
                              PerfStat.VOL,
                              PerfStat.SHARPE,
                              PerfStat.SHARPE_LOG_AN,
                              PerfStat.MAX_DD,
                              PerfStat.MAX_DD_VOL,
                              PerfStat.SKEWNESS,
                              PerfStat.KURTOSIS,
                              PerfStat.ALPHA,
                              PerfStat.BETA,
                              PerfStat.R2)

LN_BENCHMARK_TABLE_COLUMNS_SHORT = (PerfStat.TOTAL_RETURN,
                                    PerfStat.PA_RETURN,
                                    PerfStat.AN_LOG_RETURN,
                                    PerfStat.VOL,
                                    PerfStat.SHARPE,
                                    PerfStat.SHARPE_LOG_AN,
                                    PerfStat.MAX_DD,
                                    PerfStat.MAX_DD_VOL,
                                    PerfStat.SKEWNESS,
                                    PerfStat.ALPHA,
                                    PerfStat.BETA,
                                    PerfStat.R2)

EXTENDED_TABLE_COLUMNS = (PerfStat.START_DATE,
                          PerfStat.END_DATE,
                          PerfStat.START_PRICE,
                          PerfStat.END_PRICE,
                          PerfStat.TOTAL_RETURN,
                          PerfStat.PA_RETURN,
                          PerfStat.VOL,
                          PerfStat.SHARPE,
                          PerfStat.SHARPE_EXCESS,
                          PerfStat.MAX_DD,
                          PerfStat.MAX_DD_VOL,
                          PerfStat.SKEWNESS,
                          PerfStat.KURTOSIS)

COMPACT_TABLE_COLUMNS = (PerfStat.TOTAL_RETURN,
                         PerfStat.PA_RETURN,
                         PerfStat.VOL,
                         PerfStat.SHARPE,
                         PerfStat.MAX_DD,
                         PerfStat.MAX_DD_VOL,
                         PerfStat.SKEWNESS)

SMALL_TABLE_COLUMNS = (PerfStat.TOTAL_RETURN,
                       PerfStat.PA_RETURN,
                       PerfStat.VOL,
                       PerfStat.SHARPE,
                       PerfStat.MAX_DD)

BENCHMARK_TABLE_COLUMNS = (PerfStat.PA_RETURN,
                           PerfStat.VOL,
                           PerfStat.SHARPE,
                           PerfStat.MAX_DD,
                           PerfStat.MAX_DD_VOL,
                           PerfStat.SKEWNESS,
                           PerfStat.ALPHA_AN,
                           PerfStat.BETA,
                           PerfStat.R2)


BENCHMARK_TABLE_COLUMNS2 = (PerfStat.TOTAL_RETURN,
                            PerfStat.PA_RETURN,
                            PerfStat.VOL,
                            PerfStat.SHARPE,
                            PerfStat.MAX_DD,
                            PerfStat.MAX_DD_VOL,
                            PerfStat.SKEWNESS,
                            PerfStat.ALPHA_AN,
                            PerfStat.BETA,
                            PerfStat.R2)


def compute_performance_table(prices: Union[pd.DataFrame, pd.Series],
                              perf_params: PerfParams,
                              ) -> pd.DataFrame:
    """
    get performances for assets in prices
    """
    if not isinstance(prices, pd.DataFrame):
        raise TypeError(f"must be pd.Dataframe")

    if perf_params is None:
        perf_params = PerfParams()

    dict_data = OrderedDict()
    for asset in prices:
        asset_data = prices[asset].dropna()  # force drop na
        return_dict = ret.compute_returns_dict(prices=asset_data,
                                               perf_params=perf_params)
        dict_data[asset] = return_dict

    # keys will be rows = asset, column = keys in return_dict
    data = pd.DataFrame.from_dict(data=dict_data, orient='index')
    return data


def compute_risk_table(prices: pd.DataFrame,
                       perf_params: PerfParams = None,
                       is_compute_vol_only: bool = False
                       ) -> pd.DataFrame:
    """
    compute price data for returns statistics
    prices is limited to [sample_date_0,...,sample_date_N]
    """
    if not isinstance(prices, pd.DataFrame):
        raise TypeError(f"prices must be dataframe.")

    if perf_params is None:
        perf_params = PerfParams()

    sampled_prices = ret.prices_at_freq(prices=prices, freq=perf_params.freq_vol)

    # drawdowns are computed using its own freq
    if perf_params.freq_vol == perf_params.freq_drawdown:
        dd_sampled_prices = sampled_prices
    else:
        dd_sampled_prices = ret.prices_at_freq(prices=prices, freq=perf_params.freq_drawdown)

    vol_dt = np.sqrt(da.infer_an_from_data(data=sampled_prices))
    dict_data = OrderedDict()
    for asset in sampled_prices:
        sampled_price = sampled_prices[asset].dropna()
        if len(sampled_price.index) > 1:
            sampled_returns = ret.to_returns(prices=sampled_price, return_type=perf_params.return_type, drop_first=True)
            nd_sampled_returns = sampled_returns.to_numpy()
            vol = vol_dt * np.std(nd_sampled_returns, ddof=1)

            asset_dict = {PerfStat.VOL.to_str(): vol,
                          PerfStat.AVG_LOG_RETURN.to_str(): np.nanmean(sampled_returns),
                          PerfStat.START_DATE.to_str(): sampled_price.index[0],
                          PerfStat.END_DATE.to_str(): sampled_price.index[-1],
                          PerfStat.NUM_OBS.to_str(): len(nd_sampled_returns)
                          }
            if not is_compute_vol_only:
                # compute max dd on business day schedule
                max_dd = spd.compute_max_dd(prices=dd_sampled_prices[asset].dropna())
                rel_returns = sampled_price.pct_change().dropna()
                asset_dict.update({
                    PerfStat.MAX_DD.to_str(): max_dd,
                    PerfStat.MAX_DD_VOL.to_str(): max_dd / vol if vol > 0.0 else 0.0,
                    PerfStat.WORST.to_str(): np.min(rel_returns),
                    PerfStat.BEST.to_str(): np.max(rel_returns),
                    PerfStat.SKEWNESS.to_str(): skew(nd_sampled_returns, bias=False),
                    PerfStat.KURTOSIS.to_str(): kurtosis(nd_sampled_returns, bias=False)
                })
        else:  # no required data
            asset_dict = {PerfStat.VOL.to_str(): np.nan,
                          PerfStat.AVG_LOG_RETURN.to_str(): np.nan,
                          PerfStat.START_DATE.to_str(): np.nan,
                          PerfStat.END_DATE.to_str(): np.nan,
                          PerfStat.NUM_OBS.to_str(): 0
                          }
        dict_data[asset] = asset_dict

    # keys will be rows = asset, column = keys in return_dict
    data = pd.DataFrame.from_dict(data=dict_data, orient='index')

    return data


def compute_ra_perf_table(prices: Union[pd.DataFrame, pd.Series],
                          perf_params: PerfParams = None,
                          is_compute_sharpe_only: bool = False
                          ) -> pd.DataFrame:

    if perf_params is None:
        perf_params = PerfParams(freq=pd.infer_freq(prices.index))

    if isinstance(prices, pd.Series):
        prices = prices.to_frame()

    perf_table = compute_performance_table(prices=prices, perf_params=perf_params)

    # is we only need sharpe we only comptute vol without higher order risk
    risk_table = compute_risk_table(prices=prices,
                                    perf_params=perf_params,
                                    is_compute_vol_only=is_compute_sharpe_only)

    # get vol and compute risk adjusted performance
    vol = risk_table[PerfStat.VOL.to_str()]
    perf_table[PerfStat.SHARPE.to_str()] = perf_table[PerfStat.PA_RETURN.to_str()] / vol
    perf_table[PerfStat.SHARPE_LOG_AN.to_str()] = perf_table[PerfStat.AN_LOG_RETURN.to_str()] / vol
    perf_table[PerfStat.SHARPE_AVG.to_str()] = perf_table[PerfStat.AVG_AN_RETURN.to_str()] / vol  # computed in risk
    perf_table[PerfStat.SHARPE_EXCESS.to_str()] = perf_table[PerfStat.PA_EXCESS_RETURN.to_str()] / vol

    # merge the two meta on dates
    ra_perf_table = pd.merge(left=perf_table, right=risk_table,
                             left_index=True, right_index=True, how='inner',
                             suffixes=(None, "_y"))
    return ra_perf_table


def compute_ra_perf_table_with_benchmark(prices: pd.DataFrame,
                                         benchmark: str,
                                         perf_params: PerfParams = None,
                                         is_log_returns: bool = False,
                                         **kwargs
                                         ) -> pd.DataFrame:
    assert benchmark in prices.columns
    if perf_params is None:
        perf_params = PerfParams(freq=pd.infer_freq(prices.index))

    ra_perf_table = compute_ra_perf_table(prices=prices, perf_params=perf_params)

    returns = ret.to_returns(prices=prices, freq=perf_params.freq_reg, is_log_returns=is_log_returns)
    alphas, betas, r2 = {}, {}, {}
    for column in returns.columns:
        joint_data = returns[[benchmark, column]].dropna()
        if joint_data.empty or len(joint_data.index)<2:
            alphas[column], betas[column], r2[column] = np.nan, np.nan, np.nan
        else:
            alphas[column], betas[column], r2[column] = ols.estimate_ols_alpha_beta(x=joint_data.iloc[:, 0], y=joint_data.iloc[:, 1])

    # get vol and compute risk adjusted performance
    ra_perf_table[PerfStat.ALPHA.to_str()] = pd.Series(alphas)
    ra_perf_table[PerfStat.ALPHA_AN.to_str()] = pd.Series(alphas)
    ra_perf_table[PerfStat.BETA.to_str()] = pd.Series(betas)
    ra_perf_table[PerfStat.R2.to_str()] = pd.Series(r2)

    return ra_perf_table


def compute_desc_freq_table(df: pd.DataFrame,
                            freq: str = 'A',
                            agg_func: Callable = np.sum
                            ) -> pd.DataFrame:
    """
    for time series pandas, aggregate by freq and produce descriprive data
    """
    freq_data = df.resample(freq).agg(agg_func)

    # drop na rows for all
    freq_data = freq_data.dropna(axis=0, how='any')

    data_values = freq_data.to_numpy()
    data_table = pd.DataFrame(index=freq_data.columns)
    data_table[PerfStat.AVG.to_str()] = np.nanmean(data_values, axis=0)
    data_table[PerfStat.STD.to_str()] = np.nanstd(data_values, ddof=1, axis=0)
    data_table[PerfStat.QUANT_M_1STD.to_str()] = np.nanquantile(data_values, q=0.16, axis=0)
    data_table[PerfStat.MEDIAN.to_str()] = np.mean(data_values, axis=0)
    data_table[PerfStat.QUANT_P1_STD.to_str()] = np.nanquantile(data_values, q=0.84, axis=0)

    return data_table


def compute_te_ir_errors(return_diffs: pd.DataFrame) -> Tuple[pd.Series, pd.Series]:
    """
    compute information ratio from return diffs
    """
    vol_dt = da.get_vol_an(pd.infer_freq(return_diffs.index))
    avg = np.nanmean(return_diffs, axis=0)
    vol = np.nanstd(return_diffs, axis=0, ddof=1)
    ir = vol_dt * np.divide(avg, vol, where=np.greater(vol, 0.0))
    te = pd.Series(vol_dt * vol, index=return_diffs.columns, name=PerfStat.TE.to_str())
    ir = pd.Series(ir, index=return_diffs.columns, name=PerfStat.IR.to_str())
    return te, ir


def compute_info_ratio_table(return_diffs_dict: Dict[str, pd.DataFrame]) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    compute information ratio from return diffs
    """
    te_ac_datas = []
    ir_ac_datas = []
    for ac, data in return_diffs_dict.items():
        te, ir = compute_te_ir_errors(return_diffs=data)
        te_ac_datas.append(te.rename(ac))
        ir_ac_datas.append(ir.rename(ac))
    te_table = pd.concat(te_ac_datas, axis=1)
    ir_table = pd.concat(ir_ac_datas, axis=1)
    return te_table, ir_table


class UnitTests(Enum):
    RA_PERF_TABLE = 1


def run_unit_test(unit_test: UnitTests):
    from qis.data.yf_data import load_etf_data
    prices = load_etf_data().dropna()
    print(prices)

    if unit_test == UnitTests.RA_PERF_TABLE:

        perf_params = PerfParams(freq='B')
        table = compute_ra_perf_table(prices=prices, perf_params=perf_params)
        print(table)
        print(table.columns)


if __name__ == '__main__':

    unit_test = UnitTests.RA_PERF_TABLE

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
common pandas operations
"""
# built in
import warnings
import numpy as np
import pandas as pd
from scipy import stats
from typing import Union, List, Optional, Dict, Tuple, Type, Any, Literal
from enum import Enum

# qis
import qis.utils.np_ops as npo
import qis.utils.struct_ops as sop


def df_zero_like(df: pd.DataFrame) -> pd.DataFrame:
    return pd.DataFrame(data=np.zeros(df.shape), index=df.index, columns=df.columns)


def df_ones_like(df: pd.DataFrame) -> pd.DataFrame:
    return pd.DataFrame(data=np.ones(df.shape), index=df.index, columns=df.columns)


def df_indicator_like(df: pd.DataFrame, type: Type = bool) -> pd.DataFrame:
    """
    return df_indicator = True is set is not None and False otherwise
    """
    data = np.where(np.isnan(df.to_numpy(dtype=np.float64)), False, True).astype(type)
    return pd.DataFrame(data=data, index=df.index, columns=df.columns)


def dfs_indicators(dfs: List[pd.DataFrame], type: Type = bool) -> pd.DataFrame:
    """
    return df_indicator = True if list dfs is not none
    nb: need to check that dfs are aligned
    """
    def to_indicator(df: pd.DataFrame) -> np.ndarray:
        return np.where(np.isnan(df.to_numpy(dtype=np.float64)), False, True)
    indicators = to_indicator(df=dfs[0])
    for df in dfs[1:]:
        indicators = np.logical_and(indicators, to_indicator(df=df))
    return pd.DataFrame(data=indicators, index=dfs[0].index, columns=dfs[0].columns).astype(type)


def df_filter_like(df: pd.DataFrame,
                   df_filter: pd.DataFrame
                   ) -> pd.DataFrame:
    """
    take a subset of set with data conditioned on non nans in
    """
    df_filter = df_filter[df.columns].reindex(index=df.index)
    df_out = df.where(cond=df_indicator_like(df=df_filter), other=np.nan)
    return df_out


def df_joint_indicator(indicator1: pd.DataFrame,
                       indicator2: Union[np.ndarray, pd.DataFrame],
                       to_float: bool = False
                       ) -> pd.DataFrame:
    """
    return df_indicator = True if indicator1 is True and indicator2 is True
    """
    if isinstance(indicator2, pd.DataFrame):
        np_indicator2 = indicator2.to_numpy(dtype=bool)
    elif isinstance(indicator2, np.ndarray):
        np_indicator2 = indicator2.astype(bool)
    else:
        raise ValueError(f"unsupported type {type(indicator2)}")

    data = np.logical_and(indicator1.to_numpy(dtype=bool), np_indicator2)

    if to_float:
        data = data.astype(np.float64)

    return pd.DataFrame(data=data, index=indicator1.index, columns=indicator1.columns)


def df_indicator_like_other(indicator: pd.DataFrame,
                            other: pd.DataFrame
                            ) -> pd.DataFrame:
    """
    return set = other if indicator is True and np.nan otherwise
    """
    data = np.where(indicator.to_numpy(dtype=bool), other.to_numpy(), np.nan)
    return pd.DataFrame(data=data, index=other.index, columns=other.columns)


def series_to_dict(ds: pd.Series, is_inverse: bool = False) -> Dict:
    """
    convert pd series to dict {index: value} and reverse
    important only uniue values
    """
    sop.assert_list_unique(lsdata=ds.index.to_list())
    data_dict = ds.to_dict()
    if is_inverse:
        sop.assert_list_unique(lsdata=ds.to_list())  # values need to be unique
        data_dict = {v: k for k, v in data_dict.items()}
    return data_dict


def norm_df_by_ax_mean(data: Union[np.ndarray, pd.DataFrame],
                       proportiontocut: Optional[float] = None,  # for treaming mean,
                       is_zero_mean: bool = True,
                       axis: int = 1
                       ) -> pd.DataFrame:
    """
    to unit norm data
    axis: 0: time-series row-wise mean; 1: cross-sectional columns-wise mean
    """

    if isinstance(data, pd.DataFrame):
        a = data.to_numpy()
    else:
        a = data

    if is_zero_mean:
        if proportiontocut is not None:
            mean = npo.np_array_to_df_columns(a=stats.trim_mean(a=a,
                                                                proportiontocut=proportiontocut, axis=axis),
                                              n_col=len(data.columns))
        else:
            mean = np.nanmean(a=a, axis=axis, keepdims=True)
        a = a - mean

    std = np.nanstd(a=a, axis=axis, ddof=1, keepdims=True)
    norm_data = np.divide(a, std, where=np.greater(std, 0.0))

    if isinstance(data, pd.DataFrame):
        norm_data = pd.DataFrame(data=norm_data, columns=data.columns, index=data.index)

    return norm_data


def get_first_before_non_nan_index(df: Union[pd.Series, pd.DataFrame],
                                   return_index_for_all_nans: int = -1  # last = -1, first = 0
                                   ) -> Union[pd.Timestamp, List[pd.Timestamp]]:
    """
    return first index if no nans in pandas
    """
    if isinstance(df, pd.DataFrame):
        first_before_non_nan_index = []
        for column in df:
            column_data = df[column]
            non_nan_data = column_data.loc[np.isnan(column_data.to_numpy()) == False]
            if not non_nan_data.empty:  # returns first index of nan
                fist_non_nan_index = non_nan_data.index[0]
            else:  # all values are nans so return the last index of original data
                fist_non_nan_index = df.index[return_index_for_all_nans]

            first_before_non_nan_index.append(fist_non_nan_index)

    elif isinstance(df, pd.Series):
        non_nan_data = df.loc[np.isnan(df.to_numpy()) == False]
        if not non_nan_data.empty:
            first_before_non_nan_index = non_nan_data.index[0]
        else:
            first_before_non_nan_index = df.index[return_index_for_all_nans]
    else:
        raise ValueError(f"unsoported data type = {type(df)}")

    return first_before_non_nan_index


def drop_first_nan_data(df: Union[pd.Series, pd.DataFrame],
                        is_oldest: bool = True  # the olderst, the eldest (all non nnans)
                        ) -> Union[pd.Series, pd.DataFrame]:
    """
    drop data before first nonnan either at max or at min for pandas
    the rest of data can still contain occasional nans
    """
    first_non_nan_index = get_first_before_non_nan_index(df=df)

    if isinstance(df, pd.DataFrame):
        if is_oldest:
            joint_start = min(first_non_nan_index)
        else:
            joint_start = max(first_non_nan_index)
        new_data = df[joint_start:].copy()
    else:
        new_data = df.loc[first_non_nan_index:].copy()

    return new_data


def get_first_last_non_nan_date(df: Union[pd.Series, pd.DataFrame], is_first: bool = True) -> Union[pd.Timestamp, np.ndarray]:
    """
    return pd.Timestamp or np.nan
    """
    def get_series_non_nan(ds: pd.Series) -> pd.Timestamp:
        null_ind = ds.isnull()
        good_ind = null_ind == False
        if np.all(good_ind):
            if is_first:  # no nans nans return first:
                fist_non_nan_index = ds.index[0]
            else:  # last
                fist_non_nan_index = ds.index[-1]
        elif np.all(null_ind):
            if is_first:  # all nans return last:
                fist_non_nan_index = ds.index[-1]
            else:
                fist_non_nan_index = ds.index[0]
        else:
            if is_first:
                fist_non_nan_index = ds[good_ind].index[0]
            else:
                fist_non_nan_index = ds[good_ind].index[-1]
        return fist_non_nan_index

    if isinstance(df, pd.Series):
        fist_non_nan_index = get_series_non_nan(df)
    elif isinstance(df, pd.DataFrame):
        values = []
        for column in df:
            values.append(get_series_non_nan(df[column]))
        fist_non_nan_index = np.array(values)
    else:
        raise ValueError(f"unsupported data type = {type(df)}")

    return fist_non_nan_index


def missing_zero_ratios_after_first_non_nan(df: Union[pd.Series, pd.DataFrame],
                                            zero_cutoff: float = 1e-12
                                            ) -> Tuple[np.ndarray, np.ndarray]:
    """
   compute ration of missing data
    """
    if isinstance(df, pd.Series):
        df = df.to_frame()

    missing = []
    zeros = []
    for column in df:
        column_data = df[column]
        nonnan_cond = column_data.isnull() == False
        fist_non_nan_index = column_data.loc[nonnan_cond].index[0]
        after_data = column_data.loc[fist_non_nan_index:]
        missing.append(after_data.isnull().sum() / len(after_data))
        zeros.append(after_data.abs().lt(zero_cutoff).sum() / len(after_data))

    missing_ratio = np.array(missing)
    zeros_ratio = np.array(zeros)

    return missing_ratio, zeros_ratio


def get_first_non_nan_values(df: Union[pd.Series, pd.DataFrame]) -> Union[np.ndarray, float]:

    if df.empty:
        raise ValueError(f"data is empty:\n {df}")

    def get_non_nan_values_series(sdata: pd.Series) -> float:
        x0 = sdata.iloc[0]
        if not np.isnan(x0):
            values = x0
        else:
            nonnan_column_data = sdata[~sdata.isnull()]
            if nonnan_column_data.empty:
                raise ValueError(f"all nans in {df}")
            values = nonnan_column_data[0]
        return values

    if isinstance(df, pd.DataFrame):
        x0 = df.iloc[0, :].to_numpy()
        if np.all(np.isnan(x0) == False):  # first entries are non nan -> most expected
            values = x0
        else:
            values = []
            for column in df:
                values.append(get_non_nan_values_series(sdata=df[column]))
            values = np.array(values)

    elif isinstance(df, pd.Series):
            values = get_non_nan_values_series(sdata=df)

    else:
        raise ValueError(f"unsupported data type = {type(df)}")

    return values


def get_last_non_nan_values(df: Union[pd.Series, pd.DataFrame]) -> Union[np.ndarray, float]:
    if df.empty:
        raise ValueError(f"data is empty:\n {df}")

    def get_non_nan_values_series(ds: pd.Series) -> float:
        x1 = ds.iloc[-1]
        if not np.isnan(x1):
            values = x1
        else:
            nonnan_column_data = ds[~ds.isnull()]
            if nonnan_column_data.empty:
                print(f"in get_last_non_nan_values: all nans in {ds}")
                values = np.nan
            else:
                values = nonnan_column_data[-1]
        return values

    if isinstance(df, pd.DataFrame):
        x1 = df.iloc[-1, :].to_numpy()
        if np.all(np.isnan(x1) == False):  # first entries are non nan -> most expected
            values = x1
        else:
            values = []
            for column in df:
                values.append(get_non_nan_values_series(ds=df[column]))
            values = np.array(values)

    elif isinstance(df, pd.Series):
        values = get_non_nan_values_series(ds=df)

    else:
        raise ValueError(f"unsupported data type = {type(df)}")

    return values


def get_df_subset_index(df: pd.DataFrame,
                        sub_index_data: Union[pd.Series, pd.DataFrame]
                        ) -> pd.DataFrame:
    """
    get subset of pandas based on smaller index in sub_index_data
    """
    return df.iloc[np.in1d(df.index.to_list(), sub_index_data.index.to_list())]


def multiply_df_by_dt(df: Union[pd.DataFrame, pd.Series],
                      dates: Union[pd.DatetimeIndex, pd.Index] = None,
                      lag: Optional[int] = None,
                      is_actual_calendar_dt: bool = True,
                      af: float = 365.0
                      ) -> Union[pd.DataFrame, pd.Series]:
    """
    to compute rate adjustment with data - rate:
    get data at dates index and adjust by dt if needed
    adjust data by time spread:
    data = dt*data
    """
    if not isinstance(df.index, pd.DatetimeIndex):
        raise TypeError(f"data index must be DateTimeIndex: {df.index}")

    if lag is not None:
        if len(df.index) > lag:
            df = df.shift(lag)

    if dates is not None:
        df = df.reindex(index=dates, method='ffill')

    # apply dt multiplication
    if len(df.index) > 1:
        if is_actual_calendar_dt:
            delta = np.append(0.0, (df.index[1:] - df.index[:-1]).days / af)
        else:
            delta = 1.0 / af
        df = df.multiply(delta, axis=0)
    else:
        warnings.warn(f"in adjust_data_with_dt: lengh of data index is one - cannot adjust by dt")
        return df

    return df


def rename_column_names(df: pd.DataFrame, columns_map: pd.Series) -> pd.DataFrame:
    """
    rename columns according to given map
    """
    df.columns = pd.Series(df.columns).map(columns_map).to_list()
    return df


def np_txy_tensor_to_pd_dict(np_tensor_txy: np.ndarray,
                             dateindex: Union[pd.Index, pd.DatetimeIndex],
                             factor_names: Union[pd.Index, List[str]],
                             asset_names: Union[pd.Index, List[str]],
                             is_factor_view: bool = True
                             ) -> Dict[str, pd.DataFrame]:
    """
    np_tensor_txy is an output of factor model: betas = [time, factor, asset]
    given timeindex such that #timeindex = time
    convert to:
    factor view = dict{factor: set[timeindex, asset]}  for each factor exp by assets
    asset view = dict{asset: set[timeindex, factor]}  for each asset exp by factor
    """
    if np_tensor_txy.ndim != 3:
        raise TypeError(f"np_tensor must ne 3-d tensor")
    if len(dateindex) != np_tensor_txy.shape[0]:
        raise TypeError(f"time dimension must be equal")
    if len(factor_names) != np_tensor_txy.shape[1]:
        raise TypeError(f"factor dimension must be equal")
    if len(asset_names) != np_tensor_txy.shape[2]:
        raise TypeError(f"asset dimension must be equal")

    if is_factor_view:  # {factor_id: pd.DataFrame(factor loadings)}
        factor_loadings = {}
        for factor_idx, factor in enumerate(factor_names):
            factor_loadings[factor] = pd.DataFrame(data=np_tensor_txy[:, factor_idx, :],
                                                   index=dateindex,
                                                   columns=asset_names)

    else:  # {asset_id: pd.DataFrame(factor loadings)}
        factor_loadings = {}
        for asset_idx, asset in enumerate(asset_names):
            factor_loadings[asset] = pd.DataFrame(data=np_tensor_txy[:, :, asset_idx],
                                                  index=dateindex,
                                                  columns=factor_names)
    return factor_loadings


def factor_dict_to_asset_dict(factor_loadings: Dict[str, pd.DataFrame],
                              asset_names: Optional[Union[pd.Index, List[str]]] = None  # column names in set
                              ) -> Dict[str, pd.DataFrame]:
    """
    revert the dictionary Dict[factor/model: set[assets] to Dict[assets: set factor)
    """
    if asset_names is None:
        asset_names = factor_loadings[list(factor_loadings.keys())[0]].columns

    asset_factor_loadings = {}
    for asset in asset_names:
        asset_factor_loading = []
        for factor, factor_data in factor_loadings.items():
            asset_factor_loading.append(factor_data[asset].rename(factor))
        asset_factor_loadings[asset] = pd.concat(asset_factor_loading, axis=1)

    return asset_factor_loadings


def df_time_dict_to_pd(factor_loadings_dict: Dict[pd.Timestamp, pd.DataFrame],
                       is_factor_view: bool = True,
                       is_zeros_to_nan: bool = False
                       ) -> Dict[str, pd.DataFrame]:
    """
    {time: set[factor, asset}
    convert to:
    factor view = dict{factor: set[timeindex, asset]}
    asset view = dict{asset: set[timeindex, factor]}
    """
    data0 = next(iter(factor_loadings_dict.values()))
    factor_names = data0.index
    asset_names = data0.columns
    dateindex = pd.DatetimeIndex(factor_loadings_dict.keys())

    np_tensor_txy = np.full((len(dateindex), len(factor_names), len(asset_names)), np.nan)
    for idx, data in enumerate(factor_loadings_dict.values()):
        np_tensor_txy[idx, :, :] = data.to_numpy()

    if is_zeros_to_nan:
        np_tensor_txy = np.where(np.isclose(np_tensor_txy, 0.0), np.nan, np_tensor_txy)

    factor_loadings = np_txy_tensor_to_pd_dict(np_tensor_txy=np_tensor_txy,
                                               dateindex=dateindex,
                                               factor_names=factor_names,
                                               asset_names=asset_names,
                                               is_factor_view=is_factor_view)
    return factor_loadings


def dfs_to_upper_lower_diag(df_upper: pd.DataFrame,
                            df_lower: pd.DataFrame,
                            diagonal: pd.Series
                            ) -> pd.DataFrame:
    """
    create up/low/diag copz
    nb must be square matrizes
    """
    n_rows = len(df_upper.index)
    out = df_upper.copy()
    for i in range(n_rows):
        out.iloc[i, i] = diagonal.iloc[i]
        for j in range(n_rows):
            if i < j:
                out.iloc[i, j] = df_upper.iloc[i, j]
            elif i > j:
                out.iloc[i, j] = df_lower.iloc[i, j]
            else:
                pass
    return out


def df_concat_with_tz(dfs: List[pd.DataFrame]) -> pd.DataFrame:
    """
    can only do if common tz is found
    """
    # find common tz
    tz = None
    for df in dfs:
        tz1 = df.index.tz
        if tz is not None:
            if tz1 != tz:
                raise ValueError(f"{tz1} {tz}")
            else:
                tz = tz1
        else:
            tz = tz1

    # localize
    for df in dfs:
        if df.index.tz is None:
            df.index = df.index.tz_localize(tz)

    return pd.concat(dfs, axis=1)


def compute_last_score(df: Union[pd.DataFrame, pd.Series], is_percent: bool = True) -> pd.Series:
    """
    columnwise score for last value in data
    use column wise loop with percentileofscore (as it supports only float for score)
    """
    if isinstance(df, pd.Series):
        df = df.to_frame()

    percentiles = []
    for column in df.columns:
        x = df[column].dropna()
        percentiles.append(stats.percentileofscore(a=x, score=x.iloc[-1], kind='rank'))
    percentiles = pd.Series(percentiles, index=df.columns)
    if is_percent:
        percentiles = percentiles.multiply(0.01)

    return percentiles


def align_dfs_dict_with_df(dfd: Dict[Any, pd.DataFrame],
                           df: pd.DataFrame,
                           method: str = 'ffill'
                           ) -> Dict[Any, pd.DataFrame]:
    """
    align dict of dataframes with with given df
    """
    alignment_columns = df.columns
    alignment_index = df.index
    aligned_dfd = {}
    for key, df in dfd.items():
        if df is not None:  # align with other pandas in dict by filtered value
            aligned_dfd[key] = df[alignment_columns].reindex(index=alignment_index).fillna(method=method)
        else:
            aligned_dfd[key] = None
    return aligned_dfd


def align_column_to_index(df_column: pd.DataFrame,
                          df_index: pd.DataFrame,
                          join: str = 'inner',
                          is_strict_join: bool = True
                          ) -> pd.DataFrame:
    """
    align dataframe columns to data_index
    join='inner' with strict match
    join='outer' without strict match so missing data has na
    return is df = data_index with columns matched to data_column
    important is that columns in data_column (narrower) correspond to data_index index
    """
    # 1 - make pd. Series indeexd by columns of data_column
    data_columns0 = pd.Series(data=df_column.columns, index=df_column.columns)

    # 2-align data columns to the index
    data_index1, data_columns1 = df_index.align(other=data_columns0, join=join, axis=0)

    if join == 'inner' and is_strict_join:
        if len(df_column.columns) != len(data_index1.index):
            print(df_column.columns)
            print(data_index1.index)
            raise ValueError(f"columns in data_column and index are not matched")

    return data_index1


def align_index_to_index(df1: Union[pd.DataFrame, pd.Series],
                         df2: Union[pd.DataFrame, pd.Series],
                         join: Literal['inner', 'outer', 'left', 'right'] = 'inner'
                         ) -> Tuple[Union[pd.DataFrame, pd.Series], Union[pd.DataFrame, pd.Series]]:
    """
    align dataframes
    join='inner' with strict match
    join='outer' without strict match so missing data has na
    """
    data_index1_, data_index2_ = df1.align(other=df2, join=join, axis=0)

    if join == 'inner':

        if isinstance(df1, pd.Series) and isinstance(df2, pd.Series):
            if len(df1.index) != len(data_index1_.index) \
                    or len(df2.index) != len(data_index2_.index):
                raise ValueError(f"Series1.index{df1.index} and Series2.index{df2.index} are not matched")

        elif isinstance(df1, pd.DataFrame) and isinstance(df2, pd.DataFrame):
            if len(df1.index) != len(data_index1_.index) \
                    or len(df1.columns) != len(data_index1_.columns) \
                    or len(df2.index) != len(data_index2_.index) \
                    or len(df2.columns) != len(data_index2_.columns):
                raise ValueError(f"data1{df1} and data2 index{df2}"
                                 f"or columns are not matched:")

        elif isinstance(df1, pd.DataFrame) and isinstance(df2, pd.Series):
            if len(df1.index) != len(data_index1_.index) \
                    or len(df1.columns) != len(data_index1_.columns) \
                    or len(df2.index) != len(data_index2_.index):
                raise ValueError(f"data1{df1} and data2{df2} index are not matched")

        elif isinstance(df1, pd.Series) and isinstance(df2, pd.DataFrame):
            if len(df2.index) != len(data_index2_.index) \
                    or len(df2.columns) != len(data_index2_.columns) \
                    or len(df1.index) != len(data_index1_.index):
                raise ValueError(f"data1{df1} and data2{df2} index are not matched")

        else:
            raise TypeError('unimplemented type')

    return data_index1_, data_index2_


def align_df1_to_df2(df1: pd.DataFrame,
                     df2: pd.DataFrame,
                     join: Literal['inner', 'outer', 'left', 'right'] = 'inner',
                     axis: Optional[int] = None
                     ) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    align dataframes
    join='inner' with strict match
    join='outer' without strict match so missing data has na
    """
    if isinstance(df1, pd.DataFrame) and isinstance(df2, pd.DataFrame):
        pass
    else:
        raise TypeError(f"df1 and df2 must be dataframes")

    df1_, df2_ = df1.align(other=df2, join=join, axis=axis)

    if join == 'inner':
        if len(df1.index) != len(df1_.index) or len(df2.index) != len(df2_.index):
            raise ValueError(f"df1 and df2 index are not matched: {df1.index} vs {df2.index}")

        elif len(df1.columns) != len(df1_.columns) or len(df2.columns) != len(df2_.columns):
            raise ValueError(f"df1 and df2 columns are not matched: {df1.columns} vs {df2.columns}")

    return df1_, df2_


def df12_merge_with_tz(df1: pd.DataFrame,
                       df2: pd.DataFrame,
                       tz: Optional[str] = None,
                       freq: str = 'B'
                       ) -> pd.DataFrame:
    """
    tz sensetive alignment
    """
    # localize index and convert to common time zeone
    df1 = df1.copy()
    df1.index = df1.index.tz_localize(tz=tz)
    df2 = df2.copy()
    df2.index = df2.index.tz_localize(tz=tz)

    # concat
    dfs = pd.concat([df1, df2], axis=1)
    dfs = dfs.fillna(method='ffill').asfreq(freq=freq, method='ffill')
    dfs.index = dfs.index# .tz_localize(tz=tz)
    return dfs


def merge_on_column(df1: pd.DataFrame,
                    df2: pd.DataFrame,
                    is_drop_y: bool = True
                    ) -> pd.DataFrame:
    """
    merge on homogeneous column with preservation of indices and drop dublicated, _y columns
    """
    joint_data = pd.merge(left=df1, right=df2,
                          left_index=True, right_index=True,
                          suffixes=('', '_y'), how='inner')
    if is_drop_y:
        unique_columns = sop.merge_lists_unique(list1=df1.columns.to_list(), list2=df2.columns.to_list())
        joint_data = joint_data[unique_columns]
    return joint_data


def concat_at_start_date(dfs: List[Union[pd.Series, pd.DataFrame]],
                         start_date: pd.Timestamp
                         ) -> pd.DataFrame:
    data = pd.concat(dfs, axis=1).loc[start_date:]
    return data


class UnitTests(Enum):
    ALIGN = 1
    SCORES = 2


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.ALIGN:

        desc_dict1 = {'f0': (0.5, 0.5), 'f1': (0.5, 0.5), 'f2': (0.5, 0.5), 'f3': (0.5, 0.5), 'f4': (0.5, 0.5)}
        df1 = pd.DataFrame.from_dict(desc_dict1, orient='index', columns=['as1', 'as2'])

        desc_dict2 = {'f1': (1.0, 1.0), 'f2': (1.0, 1.0), 'f5': (1.0, 1.0), 'f6': (1.0, 1.0)}
        df2 = pd.DataFrame.from_dict(desc_dict2, orient='index', columns=['as1', 'as3'])

        print(f"df1=\n{df1}")
        print(f"df2=\n{df2}")

        df1_, df2_ = align_df1_to_df2(df1=df1, df2=df2, join='outer', axis=0)
        print(f"df1_=\n{df1_}")
        print(f"df2_=\n{df2_}")

        df1_, df2_ = align_df1_to_df2(df1=df1, df2=df2, join='outer', axis=None)
        print(f"df1_=\n{df1_}")
        print(f"df2_=\n{df2_}")

    elif unit_test == UnitTests.SCORES:
        np.random.seed(1)
        nrows, ncol = 20, 5
        df = pd.DataFrame(data=np.random.normal(0.0, 1.0, size=(nrows, ncol)),
                          columns=[f"id{n + 1}" for n in range(ncol)])
        print(df)
        percentiles = compute_last_score(df=df)
        print(percentiles)


if __name__ == '__main__':

    unit_test = UnitTests.ALIGN

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
factor prediction
estimate factor model:

y_t = B^T * x_t

x_t is vector of predictors
y_t is vector of tradable factors
B is factor loading
the goal is to compute the prediction of y_t given y_t and estimated B
factors loadings B are not stored

x_t and y_t are of type vardata with defined normalization

x_t and y_t most typicall with freq='B'
model is updated as freq >= 'B' say month
the data used is daily up ast period = monthth and forecasted using daily freq for current months
"""
# built in
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from enum import Enum
from typing import NamedTuple, Optional, Tuple, Union, Dict, Any

# sklearn
import xgboost as xgb
from sklearn.linear_model import ElasticNet, Lasso, Ridge, MultiTaskLasso
from sklearn.multioutput import MultiOutputRegressor
from group_lasso import GroupLasso

# qis
import qis.utils.np_ops as npo
import qis.utils.sampling as sut
import qis.models.ml.var_data as vd
import qis.models.linear.ewm as ewm
from qis.models.ml.var_data import VarData


class RegModel(Enum):
    LASSO = 1
    MULTI_TASK_LASSO = 2
    GROUP_LASSO = 3
    RIDGE = 4
    ELASTIC_NET = 5
    XGBOOST = 6


class PredictionType(Enum):
    SIMPLE = 1
    Y_WITH_ALPHA = 2


def estimate_y_prediction(x_data: pd.DataFrame,  # factor
                          y_data: pd.DataFrame,  # yvars variable
                          alpha: float = 0.001,
                          test_period_x: Union[pd.DataFrame, np.ndarray] = None,
                          reg_model: RegModel = RegModel.LASSO,
                          fit_intercept: bool = True,
                          estimation_ewm_lambda: Optional[float] = 0.7,
                          is_output_params: bool = False,
                          xgboost_dict: Dict[str, Any] = None
                          ) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
    """
    estimate time series ml y_{t} = beta * x_{t}
    estimation_ewm_lambda defines weithing of the ml
    if test data is pased it is assumed that test_period_x_{t} is updated obesrvations for predicition of y
    prediction is not weighted with lambda
    """
    X = x_data.to_numpy()
    y = y_data.to_numpy()

    if y.ndim == 1:  # make it column array, -1 is refered as the num of elements tin original row vector
        y = y.reshape((-1, ))
    else:  # maltivar display is not possible
        is_display_stats = False

    if reg_model == RegModel.LASSO:
        model = Lasso(alpha=alpha, fit_intercept=fit_intercept)

    elif reg_model == RegModel.MULTI_TASK_LASSO:
        model = MultiTaskLasso(alpha=alpha, fit_intercept=fit_intercept)

    elif reg_model == RegModel.GROUP_LASSO:
        model = GroupLasso(groups=None,
                           group_reg=5,
                           l1_reg=alpha,
                           frobenius_lipschitz=True,
                           scale_reg="inverse_group_size",
                           subsampling_scheme=1,
                           supress_warning=True,
                           n_iter=1000,
                           tol=1e-3)

    elif reg_model == RegModel.RIDGE:
        model = Ridge(alpha=alpha, fit_intercept=fit_intercept)

    elif reg_model == RegModel.ELASTIC_NET:
        model = ElasticNet(alpha=alpha, fit_intercept=fit_intercept)

    elif reg_model == RegModel.XGBOOST:
        base_model = xgb.XGBRegressor(**xgboost_dict)
        model = MultiOutputRegressor(base_model)
    else:
        raise ValueError(f"{reg_model} not implemented")

    if estimation_ewm_lambda is not None:
        weights = npo.compute_expanding_power(n=len(x_data.index),
                                              power_lambda=np.sqrt(estimation_ewm_lambda),
                                              is_reverse=True)
        X = np.dot(np.diag(weights), X)
        y = np.dot(np.diag(weights), y)

    model.fit(X, y)
    if is_output_params:
        # shape of params = (len(y.columns), len(x.columns)
        # transpose to report as factor * assets
        if reg_model == RegModel.XGBOOST:
            params = 0.0
        elif reg_model == RegModel.GROUP_LASSO:
            params = model.coef_
        else:
            params = np.transpose(model.coef_)
        output_params = pd.DataFrame(data=params, index=x_data.columns, columns=y_data.columns)
    else:
        output_params = None

    if isinstance(test_period_x, pd.DataFrame):
        xtest = test_period_x.to_numpy()
        ypred = model.predict(xtest)
        ypred = pd.DataFrame(data=ypred, index=test_period_x.index, columns=y_data.columns)

    elif isinstance(test_period_x, np.ndarray):
        if test_period_x.ndim == 1:  # nd.array need to be transposed to row arras
            if test_period_x is None:
                xtest = X[-1].reshape((1, -1))  #  take last value (single sample) in x and reshape to row array
            else:
                xtest = test_period_x.reshape((1, -1))
        else:
            xtest = test_period_x
        ypred = model.predict(xtest)

    else:
        ypred = None

    return ypred, output_params


class RegressionPredictor:

    class MethodParams(NamedTuple):
        model_update_freq: str = 'M'
        data_freq: Optional[str] = 'B' # default
        alpha: float = 0.001
        fit_intercept: bool = False
        x_lag: Optional[int] = 1  # defines lag estimation
        roll_period: int = 5*12 # defines sample period for model estimation
        reg_model: RegModel = RegModel.LASSO
        prediction_ewm_lambda: Optional[float] = 0.90
        xgboost_dict: Dict[str, Any] = None

    def __init__(self,
                 x_data: VarData,
                 y_data: VarData,
                 method_params: MethodParams):

        vd.validate_vardata(data=x_data)
        vd.validate_vardata(data=y_data)

        # todo: ensure alignment of X and Y
        self.X = x_data.to_var_data()
        self.Y = y_data.to_var_data()
        self.method_params = method_params

    def compute_prediction(self,
                           fit_intercept: Optional[bool] = None,
                           is_output_params: bool = False
                           ) -> Tuple[pd.DataFrame, Optional[Dict[pd.Timestamp, pd.DataFrame]]]:
        """
        estimate ml at given freq
        """
        if fit_intercept is None:
            fit_intercept = self.method_params.fit_intercept

        x_lag = self.method_params.x_lag

        train_live_samples = sut.split_to_train_live_samples(ts_index=self.X.index,
                                                             model_update_freq=self.method_params.model_update_freq,
                                                             roll_period=self.method_params.roll_period)

        agg_x = self.X
        agg_y = self.Y

        ypred_pds = []
        if is_output_params:
            output_params_dict = {}
        else:
            output_params_dict = None

        for date, validation_period in train_live_samples.train_live_dates.items():
            x_data = validation_period.train.locate(agg_x)
            y_data = validation_period.train.locate(agg_y)
            test_period_x = validation_period.live.locate(agg_x)

            if x_lag is not None and x_lag > 0:  # shift backward x var
                x_data = x_data.shift(x_lag).iloc[x_lag:, :]
                y_data = y_data.iloc[x_lag:, :]

            ypred, output_params = estimate_y_prediction(x_data=x_data,
                                                         y_data=y_data,
                                                         alpha=self.method_params.alpha,
                                                         fit_intercept=fit_intercept,
                                                         test_period_x=test_period_x,
                                                         reg_model=self.method_params.reg_model,
                                                         estimation_ewm_lambda=self.method_params.prediction_ewm_lambda,
                                                         is_output_params=is_output_params,
                                                         xgboost_dict=self.method_params.xgboost_dict)
            ypred_pds.append(ypred)
            if is_output_params:
                output_params_dict[date] = output_params

        y_prediction = pd.concat(ypred_pds, axis=0)
        y_prediction = y_prediction.loc[~y_prediction.index.duplicated(keep='last')]
        y_prediction = y_prediction.reindex(index=self.Y.index, method='ffill') # just in case

        return y_prediction, output_params_dict

    def compute_alpha_prediction(self,
                                 ewm_lambda: float = 0.94,
                                 prediction_type: PredictionType = PredictionType.Y_WITH_ALPHA,
                                 fit_intercept: bool = False,
                                 is_output_params: bool = False
                                 ) -> Tuple[pd.DataFrame, pd.DataFrame, Optional[Dict[pd.Timestamp, pd.DataFrame]]]:

        y_prediction, output_params_dict = self.compute_prediction(fit_intercept=fit_intercept,
                                                                   is_output_params=is_output_params)

        if prediction_type == PredictionType.SIMPLE:
            alpha_prediction = y_prediction
            ewm_r2 = alpha_prediction
        elif prediction_type == PredictionType.Y_WITH_ALPHA:
            alpha_prediction, ewm_r2 = ewm.compute_ewm_alpha_r2(y_data=self.Y, y_prediction=y_prediction, ewm_lambda=ewm_lambda)
            alpha_prediction = alpha_prediction.add(y_prediction)

        else:
            raise TypeError(f"not implemented")

        return alpha_prediction, ewm_r2, output_params_dict


class UnitTests(Enum):
    TEST = 1


def run_unit_test(unit_test: UnitTests):

    # data
    from qis.data.yf_data import load_etf_data
    prices = load_etf_data().dropna()
    returns = prices.asfreq('W-WED', method='ffill').pct_change()

    if unit_test == UnitTests.TEST:

        # define ml model
        x_data = VarData(data=returns.iloc[:, 0].to_frame(),
                         data_type=vd.DataType.RETURN,
                         normalization_type=vd.NormalizationType.INVERSE_VOL,
                         smoothing_type=vd.SmoothingType.EWMA)

        y_data = VarData(data=returns.iloc[:, 1:],
                         data_type=vd.DataType.RETURN,
                         normalization_type=vd.NormalizationType.INVERSE_VOL,
                         smoothing_type=vd.SmoothingType.EWMA)

        method_params = RegressionPredictor.MethodParams()
        regression_predictor = RegressionPredictor(x_data=x_data,
                                                   y_data=y_data,
                                                   method_params=method_params)

        print(regression_predictor.X)
        print(regression_predictor.Y)

        y_prediction = regression_predictor.compute_prediction()
        print(y_prediction)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.TEST

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



# built in
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from typing import List, Union, Tuple, Optional, Dict
from pandas.core.indexes.base import Index
from enum import Enum

# qis
import qis.plots.utils as put
import qis.utils.struct_ops as sop
from qis.plots.utils import LegendLineType, TrendLine, LastLabel
from qis.utils.desc_table import compute_desc_table, DescTableType


def plot_time_series(df: Union[pd.Series, pd.DataFrame],
                     is_use_sns_lineplot: bool = True,
                     linestyles: List[str] = None,
                     linewidth: float = 1.0,
                     x_date_freq: Union[str, None] = 'A',
                     date_format: str = '%b-%y',
                     legend_title: str = None,
                     legend_loc: Optional[Union[str, bool]] = 'upper left',
                     last_label: LastLabel = LastLabel.NONE,
                     sort_by_value_stretch_factor: float = 0.5,
                     trend_line: TrendLine = TrendLine.NONE,
                     legend_line_type: LegendLineType = LegendLineType.AVG_LAST,
                     desc_table_type: DescTableType = DescTableType.NONE,
                     legend_labels: List[str] = None,
                     indices_for_shaded_areas: Dict[str, Tuple[int, int]] = None,
                     x_labels: List[str] = None,
                     xlabel: str = None,
                     ylabel: str = None,
                     var_format: Optional[str] = '{:,.0f}',
                     markers: List[str] = False,
                     title: Union[str, bool] = None,
                     fontsize: int = 10,
                     markersize: int = None,
                     colors: List[str] = None,
                     first_color_fixed: bool = False,
                     last_color_fixed: bool = False,
                     x_limits: Tuple[Union[float, None], Union[float, None]] = None,
                     y_limits: Tuple[Optional[float], Optional[float]] = None,
                     is_log: bool = False,
                     ax: plt.Subplot = None,
                     **kwargs
                     ) -> plt.Figure:

    if trend_line in [TrendLine.AVERAGE, TrendLine.AVERAGE_SHADOWS, TrendLine.ZERO_SHADOWS,
                      TrendLine.TREND_LINE, TrendLine.TREND_LINE_SHADOWS]:  # these features are only available with sns
        is_use_sns_lineplot = True

    data1 = df.copy()
    if isinstance(data1, pd.DataFrame):
        pass
    elif isinstance(data1, pd.Series):
        data1 = data1.to_frame()
    else:
        raise TypeError(f"unsuported data type {type(data1)}")
    columns = data1.columns

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if colors is None:
        colors = put.get_n_colors(n=len(columns),
                                  first_color_fixed=first_color_fixed,
                                  last_color_fixed=last_color_fixed,
                                  **kwargs)

    if is_use_sns_lineplot:
        sns.lineplot(data=data1, palette=colors, dashes=False, markers=markers,
                     markersize=markersize,
                     linewidth=linewidth, ax=ax)
    else:
        data1.plot(ax=ax, color=colors, linewidth=linewidth)

    if trend_line == TrendLine.ZERO_SHADOWS:
        for column, color in zip(columns, colors):
            x0 = data1.index[0]
            x1 = data1.index[-1]
            y0 = 0
            y1 = 0
            x = [x0, x1]
            y = [y0, y1]
            ax.plot(x, y,
                    color='black',
                    linestyle='-',
                    transform=ax.transData,
                    lw=linewidth)

            x_ = (data1.index - data1.index[0]).days
            slope = (y1 - y0) / (x_[-1] - x_[0])
            y_line = [slope * x + y0 for x in x_]
            y = data1[column]
            ax.fill_between(data1.index, y, y_line, where=y_line >= y,
                            facecolor=color, interpolate=True, alpha=0.2, lw=linewidth)

    elif trend_line in [TrendLine.AVERAGE, TrendLine.AVERAGE_SHADOWS]:
        for column, color in zip(columns, colors):
            data2 = data1[column].dropna()  # exclude nans from showing the average lines
            if data2.empty:  # skip this columns from ere
                continue
            else:
                average = np.nanmean(data2.to_numpy())

            x0 = data2.index[0]
            x1 = data2.index[-1]
            y0 = average
            y1 = average
            x = [x0, x1]
            y = [y0, y1]

            ax.plot(x, y,
                    color=color,
                    linestyle=':',
                    transform=ax.transData,
                    linewidth=linewidth)

            if trend_line == TrendLine.AVERAGE_SHADOWS:
                x_ = (data2.index - data2.index[0]).days
                slope = (y1 - y0) / (x_[-1] - x_[0])
                y_line = [slope * x + y0 for x in x_]
                y = data2.to_numpy()
                ax.fill_between(data2.index, y, y_line, where=y_line >= y,
                                facecolor=color, interpolate=True, alpha=0.2, lw=linewidth)

    # add last labels
    if last_label in [LastLabel.AVERAGE_VALUE, TrendLine.AVERAGE, TrendLine.AVERAGE_SHADOWS]:
        average_dict = {}
        for column, color in zip(columns, colors):
            data2 = data1[column].dropna()  # exclude nans from showing the average lines

            if data2.empty:  # skip this columns from ere
                continue
            else:
                average = np.nanmean(data2.to_numpy())

            if last_label== LastLabel.AVERAGE_VALUE:
                if var_format is not None:
                    average_str = var_format.format(average)
                else:
                    average_str = average
                y1 = average
                variable_label = f"{column}, average={average_str}"
                average_dict.update({y1: [variable_label, color]})

        if last_label in [LastLabel.AVERAGE_VALUE, LastLabel.AVERAGE_VALUE_SORTED]:
            x1 = data1.index[-1]
            ymin, ymax = ax.get_ylim()
            mid = sort_by_value_stretch_factor * (ymax - ymin)
            if last_label == LastLabel.AVERAGE_VALUE_SORTED:
                pivot_dict = sorted(average_dict)
                locs = np.linspace(pivot_dict[0], sort_by_value_stretch_factor*mid, len(pivot_dict), endpoint=True)

            else:
                pivot_dict = average_dict
                locs = [dict for dict in pivot_dict]

            for key, loc in zip(pivot_dict, locs):
                ax.annotate(average_dict[key][0],
                            xy=(x1, key), xytext=(x1, loc),
                            fontsize=fontsize, weight='normal', color=average_dict[key][1],
                            textcoords='data', ha='left', va='bottom',
                            bbox={'boxstyle': 'round,pad=0.5', 'fc': average_dict[key][1], 'alpha': 0.1},
                            arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0'))

    if last_label != LastLabel.NONE or trend_line in [TrendLine.TREND_LINE, TrendLine.TREND_LINE_SHADOWS]:
        last_dict = {}
        for column, color in zip(columns, colors):
            y = data1[column].dropna()
            if last_label == LastLabel.LAST_VALUE:
                if var_format is not None:
                    if len(y.index) > 0:
                        variable_str = var_format.format(y.iloc[-1])
                    else:
                        variable_str = 'nan'
                else:
                    variable_str = y.iloc[-1]
                variable_label = f"{column}, last = {variable_str}"

                if len(y.index) > 0:
                    y1 = y.iloc[-1]
                else:
                    y1 = np.nan
                last_dict.update({y1: [variable_label, color]})

            if trend_line in [TrendLine.TREND_LINE, TrendLine.TREND_LINE_SHADOWS]:
                x0 = y.first_valid_index() or y.index[0] # if all are nons
                x1 = y.index[-1]
                y0 = y[x0]
                y1 = y[-1]

                x = [x0, x1]
                y = [y0, y1]
                ax.plot(x, y,
                        color=color,
                        linestyle='--',
                        transform=ax.transData,
                        linewidth=linewidth)

                if trend_line == TrendLine.TREND_LINE_SHADOWS:
                    x_ = (data1.index - data1.index[0]).days
                    slope = (y1 - y0) / (x_[-1] - x_[0])
                    y_line = [slope * x + y0 for x in x_]
                    y = data1[column]
                    ax.fill_between(data1.index, y, y_line, where=y_line >= y,
                                    facecolor=color, interpolate=True, alpha=0.2, lw=linewidth)

        # plot dicts sorted by last value
        x1 = data1.index[-1]
        ymin, ymax = ax.get_ylim()
        mid = sort_by_value_stretch_factor * (ymax - ymin)

        if last_label== LastLabel.LAST_VALUE_SORTED:
            pivot_dict = sorted(last_dict)
            locs = np.linspace(pivot_dict[0], sort_by_value_stretch_factor*mid, len(pivot_dict), endpoint=True)
        else:
            pivot_dict = last_dict
            locs = [dict for dict in pivot_dict]

        for key, loc in zip(pivot_dict,locs):
            ax.annotate(last_dict[key][0],
                        xy=(x1, key), xytext=(x1, loc),
                        fontsize=fontsize, weight ='normal', color = last_dict[key][1],
                        textcoords='data', ha='left', va='bottom',
                        bbox={'boxstyle': 'round,pad=0.5', 'fc': last_dict[key][1], 'alpha': 0.1},
                        arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0'))

    if indices_for_shaded_areas is not None:
        for col, indxs in indices_for_shaded_areas.items():
            y0 = data1.iloc[:, indxs[0]]
            y1 = data1.iloc[:, indxs[1]]
            ax.fill_between(data1.index, y0, y1, where=y0 >= y1,
                        facecolor=col, alpha=0.2,
                        interpolate=True)

    if legend_loc is not None:
        if legend_labels is None:
            if desc_table_type != DescTableType.NONE:  # use get_legend_with_stats_table
                stats_table = compute_desc_table(df=data1,
                                                 desc_table_type=desc_table_type,
                                                 var_format=var_format)
                put.set_legend_with_stats_table(stats_table=stats_table,
                                                ax=ax,
                                                colors=colors,
                                                legend_loc=legend_loc,
                                                fontsize=fontsize,
                                                **kwargs)
            else:  # generate legend_labels
                legend_labels = put.get_legend_lines(data=data1, legend_line_type=legend_line_type, var_format=var_format)

        if legend_labels is not None:
            put.set_legend(ax=ax,
                           labels=legend_labels,
                           colors=colors,
                           legend_loc=legend_loc,
                           legend_title=legend_title,
                           fontsize=fontsize,
                           **kwargs)

    else:
        ax.legend().set_visible(False)

    if linestyles is not None:
        put.set_linestyles(ax=ax, linestyles=linestyles)

    if x_date_freq is not None and isinstance(data1.index, pd.DatetimeIndex):
        put.set_date_on_axis(data=data1,
                             ax=ax,
                             x_date_freq=x_date_freq,
                             date_format=date_format,
                             fontsize=fontsize,
                             **kwargs)

    if y_limits is not None:
        put.set_y_limits(ax=ax, y_limits=y_limits)

    if x_limits is not None:
        put.set_x_limits(ax=ax, x_limits=x_limits)

    if var_format is not None:
        put.set_ax_ticks_format(ax=ax, fontsize=fontsize, xvar_format=None, yvar_format=var_format, **kwargs)

    if isinstance(data1.index, pd.DatetimeIndex):
        put.set_ax_tick_labels(ax=ax, skip_y_axis=True, fontsize=fontsize, **kwargs)

    elif isinstance(data1.index, Index):
        if x_labels is None:
            put.set_ax_ticks_format(ax=ax, fontsize=fontsize, **kwargs)

    put.set_ax_xy_labels(ax=ax, xlabel=xlabel, ylabel=ylabel, fontsize=fontsize, **kwargs)
    put.set_spines(ax=ax, **kwargs)

    if is_log:
        ax.set_yscale('log')

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize, **kwargs)

    return fig


def plot_time_series_2ax(df1: Union[pd.Series, pd.DataFrame],
                         df2: Union[pd.Series, pd.DataFrame],
                         legend_loc: Optional[str] = 'upper left',
                         legend_line_type: LegendLineType = LegendLineType.NONE,
                         legend_line_type2: LegendLineType = LegendLineType.NONE,
                         first_color_fixed: bool = False,
                         last_color_fixed: bool = False,
                         title: Optional[str] = None,
                         var_format: str = '{:,.0f}',
                         var_format_yax2: str = '{:,.0f}',
                         legend_labels: List[str] = None,
                         linestyles: List[str] = None,
                         linestyles_ax2: List[str] = None,
                         y_limits: Tuple[Optional[float], Optional[float]] = None,
                         y_limits_ax2: Tuple[Optional[float], Optional[float]] = None,
                         trend_line1: put.TrendLine = put.TrendLine.NONE,
                         trend_line2: put.TrendLine = put.TrendLine.NONE,
                         yvar_major_ticks1: np.ndarray = None,
                         yvar_major_ticks2: np.ndarray = None,
                         colors: List[str] = None,
                         fontsize: int = 10,
                         is_logs: Tuple[bool, bool] = (False, False),
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> plt.Figure:
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if isinstance(df1, pd.Series):
        df1 = df1.to_frame()
    if isinstance(df2, pd.Series):
        df2 = df2.to_frame()

    ncol1 = len(df1.columns)
    ncol2 = len(df2.columns)
    if colors is None:
        colors = put.get_n_colors(n=ncol1 + ncol2, first_color_fixed=first_color_fixed, last_color_fixed=last_color_fixed)

    ax_twin = ax.twinx()
    plot_time_series(df=df1,
                     legend_loc=None,
                     colors=colors[:ncol1],
                     var_format=None,
                     linestyles=linestyles,
                     trend_line=trend_line1,
                     is_log=is_logs[0],
                     y_limits=y_limits,
                     ax=ax,
                     **kwargs)

    plot_time_series(df=df2,
                     legend_loc=None,
                     colors=colors[ncol1:],
                     var_format=None,
                     linestyles=linestyles_ax2,
                     trend_line=trend_line2,
                     is_log=is_logs[1],
                     y_limits=y_limits_ax2,
                     ax=ax_twin,
                     **kwargs)

    put.set_ax_ticks_format(ax=ax, fontsize=fontsize, xvar_format=None, yvar_format=var_format, set_ticks=False,
                            yvar_major_ticks=yvar_major_ticks1, **kwargs)
    put.set_ax_ticks_format(ax=ax_twin, fontsize=fontsize, xvar_format=None, yvar_format=var_format_yax2, set_ticks=False,
                            yvar_major_ticks=yvar_major_ticks2, **kwargs)

    ax.tick_params(axis='x', which='both', bottom=False)

    if legend_loc is not None:
        if legend_labels is None:
            df1.columns = [f"{x} (left)" for x in df1.columns]
            df2.columns = [f"{x} (right)" for x in df2.columns]
            legend_labels1 = put.get_legend_lines(data=df1,
                                                  legend_line_type=legend_line_type,
                                                  var_format=var_format)
            legend_labels2 = put.get_legend_lines(data=df2,
                                                  legend_line_type=legend_line_type2,
                                                  var_format=var_format_yax2)
            legend_labels = sop.to_flat_list(legend_labels1 + legend_labels2)
        put.set_legend(ax=ax,
                       labels=legend_labels,
                       colors=colors,
                       fontsize=fontsize,
                       legend_loc=legend_loc,
                       **kwargs)

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize)

    return fig


def plot_lines_list(xy_datas: List[pd.DataFrame],
                    data_labels: List[List[str]],
                    colors: List[str] = None,
                    is_fill_annotations: bool = False,
                    xvar_format: str = '{:.1f}',
                    yvar_format: str = '{:.1f}',
                    xlabel: str = None,
                    ylabel: str = None,
                    title: str = None,
                    fontsize: int = 10,
                    x_limits: Tuple[Union[float, None], Union[float, None]] = None,
                    y_limits: Tuple[Union[float, None], Union[float, None]] = None,
                    ax: plt.Subplot = None,
                    **kwargs
                    ) -> plt.Figure:
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if colors is None:
        colors = put.get_n_colors(n=len(xy_datas))

    for xy, color, labels in zip(xy_datas, colors, data_labels):
        sns.lineplot(x=xy.columns[0], y=xy.columns[1], data=xy, marker='None', color=color, ax=ax)

        for label, x, y in zip(labels, xy.iloc[:, 0].to_numpy(), xy.iloc[:, 1].to_numpy()):

            if label != '':
                # very specific to blended portfolios
                port_weight = label[-4:]
                if port_weight == '100%':
                    x_offset = 40  # 100 portfolios are shifted to right
                    y_offset = 0
                elif port_weight in ['-50%', ' 60%']:
                    x_offset = 40
                    y_offset = 15
                elif port_weight == '0-0%':
                    x_offset = 40
                    y_offset = 30
                else:
                    x_offset = 1
                    y_offset = 1

                if is_fill_annotations:
                    ax.annotate(
                        label,
                        xy=(x, y), xytext=(x_offset, y_offset),
                        textcoords='offset points', ha='right', va='bottom',
                        bbox=dict(boxstyle='round, pad=0.5', fc='blue', alpha=0.75),
                        arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0'),
                        fontsize=fontsize)
                else:
                    ax.annotate(
                        label,
                        xy=(x, y), xytext=(x_offset, y_offset),
                        textcoords='offset points', ha='right', va='bottom',
                        fontsize=fontsize)

    if y_limits is not None:
        put.set_y_limits(ax=ax, y_limits=y_limits)

    if x_limits is not None:
        put.set_x_limits(ax=ax, x_limits=x_limits)

    put.set_ax_ticks_format(ax=ax, fontsize=fontsize, xvar_format=xvar_format, yvar_format=yvar_format)
    put.set_ax_xy_labels(ax=ax, xlabel=xlabel, ylabel=ylabel, fontsize=fontsize, **kwargs)

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize)

    put.set_spines(ax=ax, **kwargs)

    return fig


class UnitTests(Enum):
    PRICES = 1
    PRICES_2AX = 2


def run_unit_test(unit_test: UnitTests):

    from qis.data.yf_data import load_etf_data
    prices = load_etf_data().dropna()

    if unit_test == UnitTests.PRICES:
        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
        global_kwargs = {'fontsize': 8,
                         'linewidth': 0.5,
                         'weight': 'normal',
                         'markersize': 1}
        plot_time_series(df=prices,
                         legend_line_type=put.LegendLineType.AVG_LAST,
                         ax=ax,
                         trend_line=TrendLine.AVERAGE_SHADOWS,
                         **global_kwargs)

    elif unit_test == UnitTests.PRICES_2AX:
        fig, ax = plt.subplots(1, 1, figsize=(4, 3))
        global_kwargs = {'fontsize': 6,
                         'linewidth': 0.5,
                         'weight': 'normal',
                         'markersize': 1}

        plot_time_series_2ax(df1=prices.iloc[:, -1],
                             df2=prices.iloc[:, :-1],
                             legend_line_type=put.LegendLineType.AVG_LAST,
                             var_format_yax2='{:.0f}',
                             ax=ax,
                             **global_kwargs)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PRICES

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



from enum import Enum


class DotDict(dict):
    """
    dot.notation access to dictionary attributes
    """
    def __getattr__(self, attr):
        return self.get(attr)
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __getstate__(self):
        return self

    def __setstate__(self, state):
        self.update(state)
        self.__dict__ = self

    # def __setattr__(self, name, value):
    #    self._data[name] = value


class UnitTests(Enum):
    DOT_DICT = 1


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.DOT_DICT:
        this = DotDict({'me': 3, 'you': 10})
        print(this)
        print(this.me)
        print(this.you)

        for k, v in this.items():
            print(f"{k}: {v}")

        this['me1'] = 6
        this.me2 = 12
        for k, v in this.items():
            print(f"{k}: {v}")


if __name__ == '__main__':

    unit_test = UnitTests.DOT_DICT

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)





"""
generate and load test data using yf library
"""

import pandas as pd
import yfinance as yf
from typing import List
from enum import Enum

import qis.file_utils as fu
from qis.utils.dates import TimePeriod


LOCAL_RESOURCE_PATH = fu.get_local_paths()['LOCAL_RESOURCE_PATH']


def fetch_prices(tickers: List[str] = ('BTC-USD', ),
                 time_period: TimePeriod = None,
                 freq: str = None
                 ) -> pd.DataFrame:

    if len(tickers) == 1:
        prices = yf.download(tickers[0], start=None, end=None)
        prices = prices.rename({'Adj Close': 'close'}, axis=1)
    else:
        prices = {}
        for ticker in tickers:
            prices[ticker] = yf.download(ticker, start=None, end=None)['Adj Close']
        prices = pd.DataFrame.from_dict(prices, orient='columns')
    if time_period is not None:
        prices = time_period.locate(prices)
    if freq is not None:
        prices = prices.resample(freq).last().fillna(method='ffill')
    return prices


def load_etf_data() -> pd.DataFrame:
    prices = fu.load_df_from_csv(file_name='etf_prices', local_path=LOCAL_RESOURCE_PATH)
    return prices


class UnitTests(Enum):
    ETF_PRICES = 1
    TEST_LOADING = 2


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.ETF_PRICES:
        prices = fetch_prices(tickers=['SPY', 'QQQ', 'EEM', 'TLT', 'IEF', 'LQD', 'HYG', 'SHY', 'GLD'])
        print(prices)
        fu.save_df_to_csv(df=prices, file_name='etf_prices', local_path=LOCAL_RESOURCE_PATH)

    elif unit_test == UnitTests.TEST_LOADING:
        prices = load_etf_data()
        print(prices)


if __name__ == '__main__':

    unit_test = UnitTests.TEST_LOADING

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
generic to generate strategy factsheet report
with comparision to 1-2 benchmarks
"""
# built in
import pandas as pd
import matplotlib.pyplot as plt
from typing import Union, List, Optional
from enum import Enum

# qis
import qis.file_utils as fu
import qis.utils.dates as da
import qis.utils.struct_ops as sop
from qis.utils.df_to_str import series_to_str

import qis.perfstats.perf_table as rpt
import qis.perfstats.returns as ret
import qis.perfstats.drawdowns as cdr
from qis.perfstats.config import PerfStat, PerfParams, RegimeData
import qis.plots.derived.returns_scatter as prs
import qis.perfstats.regime_classifier as rcl
from qis.perfstats.regime_classifier import BenchmarkReturnsQuantileRegimeSpecs
import qis.models.linear.ewm_factors as ef

# plots
import qis.plots.derived.prices as ppd
import qis.plots.derived.perf_table as ppt
import qis.plots.table as ptb
import qis.plots.derived.returns_heatmap as rhe
import qis.plots.derived.regime_scatter as drc
import qis.plots.derived.correlations as pco


FIG_SIZE = (8.3, 11.7)  # A4 for portrait


PERF_COLUMNS = (
    # PerfStat.START_DATE,
    # PerfStat.END_DATE,
    PerfStat.TOTAL_RETURN,
    PerfStat.PA_RETURN,
    PerfStat.VOL,
    PerfStat.SHARPE,
    PerfStat.BEAR_SHARPE,
    PerfStat.NORMAL_SHARPE,
    PerfStat.BULL_SHARPE,
    PerfStat.MAX_DD,
    PerfStat.MAX_DD_VOL,
    PerfStat.WORST,
    PerfStat.BEST,
    PerfStat.SKEWNESS,
    PerfStat.KURTOSIS)


PERF_PARAMS = PerfParams(freq='W-WED')
REGIME_PARAMS = BenchmarkReturnsQuantileRegimeSpecs(freq='Q')


class MultiAssetReport:

    def __init__(self,
                 prices: pd.DataFrame,
                 benchmark_prices: Union[pd.Series, pd.DataFrame],
                 perf_params: PerfParams = PERF_PARAMS,
                 regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS
                 ):

        # make sure it is consistent
        self.prices = prices
        self.benchmark_prices = benchmark_prices.reindex(index=prices.index, method='ffill')
        self.perf_params = perf_params
        self.regime_params = regime_params

    def get_prices(self, benchmark: str = None, time_period: da.TimePeriod = None) -> pd.DataFrame:
        if benchmark is not None and benchmark not in self.prices.columns:
            if isinstance(self.benchmark_prices, pd.Series):
                prices = pd.concat([self.benchmark_prices, self.prices], axis=1)
            else:
                prices = pd.concat([self.benchmark_prices[benchmark], self.prices], axis=1)
        else:
            prices = self.prices
        if time_period is not None:
            prices = time_period.locate(prices)
        return prices

    def add_regime_shadows(self, ax: plt.Subplot, regime_benchmark_str: str, data_df: pd.DataFrame) -> None:
        if isinstance(self.benchmark_prices, pd.Series):
            pivot_prices = self.benchmark_prices
        else:
            if regime_benchmark_str is None:
                regime_benchmark_str = self.benchmark_prices.columns[0]
            pivot_prices = self.benchmark_prices[regime_benchmark_str]
        pivot_prices = pivot_prices.reindex(index=data_df.index, method='ffill')
        rcl.add_bnb_regime_shadows(ax=ax,
                                   data_df=data_df,
                                   pivot_prices=pivot_prices,
                                   benchmark=regime_benchmark_str,
                                   regime_params=self.regime_params)

    def plot_ra_perf_table(self,
                           benchmark: str,
                           time_period: da.TimePeriod = None,
                           ax: plt.Subplot = None,
                           **kwargs) -> None:
        prices = self.get_prices(benchmark, time_period=time_period)
        ppt.plot_ra_perf_table_benchmark(prices=prices,
                                         benchmark=benchmark,
                                         perf_params=self.perf_params,
                                         perf_columns=rpt.BENCHMARK_TABLE_COLUMNS,
                                         title=f"RA performance table: {da.get_time_period(prices).to_str()}",
                                         rotation_for_columns_headers=0,
                                         ax=ax,
                                         **kwargs)

    def plot_ra_regime_table(self,
                             regime_benchmark_str: str = None,
                             time_period: da.TimePeriod = None,
                             perf_columns: List[PerfStat] = PERF_COLUMNS,
                             columns_title: str = 'Programs',
                             first_column_width: float = 3.5,
                             ax: plt.Subplot = None,
                             **kwargs
                             ) -> plt.Figure:

        prices = pd.concat([self.prices, self.benchmark_prices], axis=1)

        if time_period is not None:
            prices = time_period.locate(prices)

        cvar_table = rcl.compute_bnb_regimes_pa_perf_table(prices=prices,
                                                           benchmark=regime_benchmark_str,
                                                           perf_params=self.perf_params,
                                                           regime_params=self.regime_params)
        table_data = pd.DataFrame(data=prices.columns, index=cvar_table.index, columns=[columns_title])

        for perf_column in perf_columns:
            table_data[perf_column.to_str()] = series_to_str(ds=cvar_table[perf_column.to_str()],
                                                             var_format=perf_column.to_format(**kwargs))

        special_columns_colors = [(0, 'steelblue')]
        fig = ptb.plot_df_table(df=table_data,
                                first_column_width=first_column_width,
                                add_index_as_column=False,
                                index_column_name='Strategies',
                                special_columns_colors=special_columns_colors,
                                ax=ax,
                                **kwargs)
        return fig

    def plot_nav(self,
                 regime_benchmark_str: str = None,
                 var_format: str = '{:.0%}',
                 sharpe_format: str = '{:.2f}',
                 title: str = 'Cumulative performance',
                 is_log: bool = True,
                 ax: plt.Subplot = None,
                 **kwargs
                 ) -> plt.Figure:
        fig = ppd.plot_prices(prices=self.prices,
                              perf_params=self.perf_params,
                              start_to_one=True,
                              is_log=is_log,
                              var_format=var_format,
                              sharpe_format=sharpe_format,
                              title=title,
                              ax=ax,
                              **kwargs)
        self.add_regime_shadows(ax=ax, regime_benchmark_str=regime_benchmark_str, data_df=self.prices)

        return fig

    def plot_drawdowns(self,
                       regime_benchmark_str: str = None,
                       ax: plt.Subplot = None,
                       **kwargs) -> None:
        cdr.plot_drawdown(prices=self.prices, ax=ax, **kwargs)
        self.add_regime_shadows(ax=ax, regime_benchmark_str=regime_benchmark_str, data_df=self.prices)

    def plot_rolling_time_under_water(self,
                                      regime_benchmark_str: str = None,
                                      ax: plt.Subplot = None,
                                      **kwargs) -> None:
        cdr.plot_rolling_time_under_water(prices=self.prices, ax=ax, **kwargs)
        self.add_regime_shadows(ax=ax, regime_benchmark_str=regime_benchmark_str, data_df=self.prices)

    def plot_annual_returns(self,
                            heatmap_freq: str = 'A',
                            date_format: str = '%Y',
                            ax: plt.Subplot = None,
                            **kwargs
                            ) -> plt.Figure:
        local_kwargs = sop.update_kwargs(kwargs=kwargs,
                                         new_kwargs=dict(fontsize=4,
                                                         square=False,
                                                         x_rotation=90))
        fig = rhe.plot_periodic_returns_table_from_prices(prices=self.prices,
                                                          freq=heatmap_freq,
                                                          ax=ax,
                                                          title=f"Annual Returns",
                                                          date_format=date_format,
                                                          **local_kwargs)
        return fig

    def plot_corr_table(self, freq: str = 'W-WED',
                        time_period: da.TimePeriod = None,
                        ax: plt.Figure = None,
                        **kwargs
                        ) -> plt.Figure:
        local_kwargs = sop.update_kwargs(kwargs=kwargs, new_kwargs=dict(fontsize=4))
        prices = self.get_prices(time_period=time_period)
        fig = pco.plot_corr_table(prices=prices,
                                  x_rotation=90,
                                  freq=freq,
                                  title=f"Correelation {freq} returns: {da.get_time_period(prices).to_str()}",
                                  ax=ax,
                                  **local_kwargs)
        return fig

    def plot_returns_scatter(self, benchmark: str, ax: plt.Figure, **kwargs):
        local_kwargs = sop.update_kwargs(kwargs=kwargs,
                                         new_kwargs={'weight': 'bold',
                                                     # 'alpha_an_factor': 52.0,
                                                     'x_rotation': 0,
                                                     'first_color_fixed': False,
                                                     'ci': None})
        fig = prs.plot_returns_scatter(prices=self.get_prices(benchmark=benchmark),
                                       benchmark=benchmark,
                                       freq=self.perf_params.freq_reg,
                                       order=2,
                                       title=f"Scatterplot of {self.perf_params.freq_reg}-returns vs {benchmark}",
                                       ax=ax,
                                       **local_kwargs)
        return fig

    def plot_benchmark_beta(self,
                            benchmark: str,
                            freq: str = 'B',
                            ewm_lambda: float = 0.97,
                            ax: plt.Subplot = None,
                            **kwargs):
        returns = ret.to_returns(prices=self.get_prices(benchmark=benchmark), freq=freq)
        ewm_linear_model = ef.estimate_ewm_linear_model(x=returns[benchmark].to_frame(),
                                                        y=returns.drop(benchmark, axis=1),
                                                        ewm_lambda=ewm_lambda,
                                                        is_x_correlated=True)
        ewm_linear_model.plot_factor_loadings(factor=benchmark,
                                              title=f"Rolling EWM-{ewm_lambda} beta to {benchmark}",
                                              ax=ax,
                                              **kwargs)
        self.add_regime_shadows(ax=ax, regime_benchmark_str=benchmark, data_df=self.prices)

    def plot_regime_data(self,
                         benchmark: str,
                         regime_data_to_plot: RegimeData = RegimeData.REGIME_SHARPE,
                         time_period: da.TimePeriod = None,
                         var_format: Optional[str] = None,
                         is_conditional_sharpe: bool = True,
                         legend_loc: Optional[str] = 'upper center',
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> plt.Figure:
        prices = self.get_prices()
        title = f"Sharpe ratio decomposition by Strategies to {benchmark} Bear/Normal/Bull regimes"
        regime_classifier = rcl.BenchmarkReturnsQuantilesRegime(regime_params=REGIME_PARAMS)
        fig = rcl.plot_regime_data(regime_classifier=regime_classifier,
                                   prices=prices,
                                   benchmark=benchmark,
                                   is_conditional_sharpe=is_conditional_sharpe,
                                   regime_data_to_plot=regime_data_to_plot,
                                   var_format=var_format or '{:.2f}',
                                   legend_loc=legend_loc,
                                   perf_params=self.perf_params,
                                   title=title,
                                   ax=ax,
                                   **kwargs)
        return fig


def generate_multi_asset_factsheet(prices: pd.DataFrame,
                                   benchmark_prices: Union[pd.Series, pd.DataFrame] = None,
                                   benchmark: str = None,
                                   perf_params: PerfParams = PERF_PARAMS,
                                   regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                                   heatmap_freq: str = 'A',
                                   time_period: da.TimePeriod = None,
                                   file_name: str = 'multi_asset_report',
                                   **kwargs
                                   ) -> None:

    # add default benchmark str
    if benchmark is None and benchmark_prices is not None:
        if benchmark_prices is None:
            raise ValueError(f"pass either benchmark or benchmark_prices")
        else:
            if isinstance(benchmark_prices, pd.Series):
                benchmark = benchmark_prices.name
            else:
                benchmark = benchmark_prices.columns[0]
    if benchmark is not None and benchmark_prices is None:
        if benchmark in prices.columns:
            benchmark_prices = prices[benchmark]
        else:
            raise ValueError(f"benchmark must be in prices")

    # report data
    report = MultiAssetReport(prices=prices,
                              benchmark_prices=benchmark_prices,
                              perf_params=perf_params,
                              regime_params=regime_params)

    local_kwargs = dict(fontsize=5,
                        linewidth=0.5,
                        weight='normal',
                        markersize=1,
                        legend_alpha=0.75)
    kwargs = sop.update_kwargs(local_kwargs, kwargs)

    # create 5*2 figure
    fig = plt.figure(figsize=FIG_SIZE, constrained_layout=True)
    gs = fig.add_gridspec(nrows=8, ncols=4, wspace=0.0, hspace=0.0)

    ax11 = fig.add_subplot(gs[:2, :2])
    report.plot_nav(regime_benchmark_str=benchmark,
                    ax=ax11,
                    **kwargs)

    ax21 = fig.add_subplot(gs[2:4, :2])
    report.plot_drawdowns(regime_benchmark_str=benchmark,
                          ax=ax21,
                          **kwargs)

    ax31 = fig.add_subplot(gs[4:6, :2])
    report.plot_rolling_time_under_water(regime_benchmark_str=benchmark,
                                         ax=ax31,
                                         **kwargs)

    ax41 = fig.add_subplot(gs[6:8, :2])
    report.plot_benchmark_beta(benchmark=benchmark,
                               ax=ax41,
                               **kwargs)

    ax12 = fig.add_subplot(gs[0, 2:])
    report.plot_ra_perf_table(benchmark=benchmark, ax=ax12, **kwargs)

    ax22 = fig.add_subplot(gs[1, 2:])
    time_period1 = da.get_time_period_shifted_by_years(time_period=da.get_time_period(df=prices))
    report.plot_ra_perf_table(benchmark=benchmark, time_period=time_period1, ax=ax22, **kwargs)

    ax23 = fig.add_subplot(gs[2:4, 2:])
    report.plot_annual_returns(ax=ax23,
                               **kwargs)

    ax24a = fig.add_subplot(gs[4, 2])
    report.plot_corr_table(freq='W-WED', ax=ax24a, **kwargs)
    ax24b = fig.add_subplot(gs[4, 3])
    report.plot_corr_table(freq='W-WED',
                           time_period=da.get_time_period_shifted_by_years(time_period=da.get_time_period(df=prices)),
                           ax=ax24b, **kwargs)

    ax25 = fig.add_subplot(gs[5, 2:])
    report.plot_regime_data(ax=ax25, benchmark=benchmark, **kwargs)

    ax26 = fig.add_subplot(gs[6:8, 2:])
    report.plot_returns_scatter(benchmark=benchmark,
                                ax=ax26,
                                **kwargs)

    fu.figs_to_pdf(figs=[fig], file_name=f"{file_name}", orientation='landscape')


class UnitTests(Enum):
    FACTSHEET = 1


def run_unit_test(unit_test: UnitTests):

    from qis.data.yf_data import load_etf_data
    prices = load_etf_data().dropna()

    if unit_test == UnitTests.FACTSHEET:
        generate_multi_asset_factsheet(prices=prices,
                                       benchmark='SPY',
                                       heatmap_freq='A',
                                       perf_params=PERF_PARAMS,
                                       regime_params=BenchmarkReturnsQuantileRegimeSpecs(freq='Q'))

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.FACTSHEET

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
plot classification regression
"""
# built in
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Optional
from enum import Enum

# quis
import qis.plots.scatter as psc
from qis.utils.df_cut import add_quantile_classification


def plot_classification_scatterplot(data: pd.DataFrame,
                                    x_column: str,
                                    y_column: str,
                                    hue_name: str = 'hue',
                                    num_buckets: Optional[int] = None,
                                    bins: np.ndarray = np.array([-3.0, -1.5, 0.0, 1.5, 3.0]),
                                    order: int = 1,
                                    title: str = None,
                                    full_sample_order: int = 3,
                                    markersize: int = 10,
                                    xvar_format: str = '{:.2f}',
                                    yvar_format: str = '{:.2f}',
                                    fit_intercept: bool = False,
                                    ax: plt.Subplot = None,
                                    **kwargs
                                    ) -> None:
    """
    add bin classification using x_column
    """
    df, _ = add_quantile_classification(df=data, x_column=x_column, hue_name=hue_name, num_buckets=num_buckets, bins=bins)

    psc.plot_scatter(df=df,
                     x_column=x_column,
                     y_column=y_column,
                     hue=hue_name,
                     fit_intercept=fit_intercept,
                     title=title,
                     order=order,
                     full_sample_order=full_sample_order,
                     markersize=markersize,
                     xvar_format=xvar_format,
                     yvar_format=yvar_format,
                     add_universe_model_ci=False,
                     add_hue_model_label=True,
                     ax=ax,
                     **kwargs)


def get_data(is_random_beta: bool = True,
             n: int = 10000
             ) -> pd.DataFrame:

    x = np.random.normal(0.0, 1.0, n)
    eps = np.random.normal(0.0, 1.0, n)

    if is_random_beta:
        beta = np.random.normal(1.0, 1.0, n)*np.abs(x)
    else:
        beta = np.ones(n)

    y = beta*x + eps
    df = pd.concat([pd.Series(x, name='x'), pd.Series(y, name='y')], axis=1)
    df = df.sort_values(by='x', axis=0)

    return df


class UnitTests(Enum):
    SCATTER = 1


def run_unit_test(unit_test: UnitTests):

    np.random.seed(2)  # freeze seed

    df1 = get_data(n=100000)
    print(df1)

    if unit_test == UnitTests.SCATTER:
        plot_classification_scatterplot(data=df1, x_column='x', y_column='y')

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.SCATTER

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
load ust rates data
"""
import numpy as np
from enum import Enum

import pandas as pd

import qis.file_utils as fu

LOCAL_RESOURCE_PATH = fu.get_local_paths()['LOCAL_RESOURCE_PATH']
UST_FILE_NAME = 'ust_rates'


COLUMN_MAP = {"BC_1MONTH": '1m',
              "BC_3MONTH": '3m',
              "BC_6MONTH": '6m',
              "BC_1YEAR": '1y',
              "BC_2YEAR": '2y',
              "BC_3YEAR": '3y',
              "BC_5YEAR": '5y',
              "BC_7YEAR": '7y',
              "BC_10YEAR": '10y',
              "BC_20YEAR": '20y',
              "BC_30YEAR": '30y'}


def generate_ust_data(folder: str = f"{LOCAL_RESOURCE_PATH}xml"):

    from ust import save_xml, read_rates, available_years, year_now

    # save UST yield rates to local folder for selected years
    for year in available_years():
        save_xml(year, folder=folder)
    save_xml(year_now(), folder=folder, overwrite=True)
    df = read_rates(start_year=1990, end_year=2022, folder=folder)
    df = df.drop('BC_30YEARDISPLAY', axis=1).rename(COLUMN_MAP, axis=1)
    df = df.replace({0.0: np.nan})
    fu.save_df_to_csv(df=df, file_name=UST_FILE_NAME, local_path=LOCAL_RESOURCE_PATH)


def load_ust_rates() -> pd.DataFrame:
    df = fu.load_df_from_csv(file_name=UST_FILE_NAME, local_path=LOCAL_RESOURCE_PATH)
    df = df.fillna(method='ffill')
    return df


def load_ust_3m_rate() -> pd.Series:
    df = fu.load_df_from_csv(file_name=UST_FILE_NAME, local_path=LOCAL_RESOURCE_PATH)
    df = df.fillna(method='ffill')
    return df['3m'] / 100.0


class UnitTests(Enum):
    GENERATE_DATA = 1
    READ_DATA = 2


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.GENERATE_DATA:
        generate_ust_data()

    elif unit_test == UnitTests.READ_DATA:
        df = load_ust_rates()
        print(df)


if __name__ == '__main__':

    unit_test = UnitTests.GENERATE_DATA

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qis',
 'qis.data',
 'qis.examples',
 'qis.models',
 'qis.models.linear',
 'qis.models.ml',
 'qis.models.stats',
 'qis.perfstats',
 'qis.plots',
 'qis.plots.derived',
 'qis.portfolio',
 'qis.portfolio.optimization',
 'qis.portfolio.reports',
 'qis.portfolio.strats',
 'qis.utils',
 'qis.utils.generic']

package_data = \
{'': ['*'], 'qis': ['resources/*', 'resources/xml/*']}

install_requires = \
['matplotlib>=3.5.2',
 'numba>=0.55',
 'numpy>=1.22.4',
 'pandas>=0.19',
 'scipy>=1.3',
 'seaborn>=0.11.2',
 'statsmodels>=0.13.0']

setup_kwargs = {
    'name': 'qis',
    'version': '1.0.1',
    'description': 'Implementation of visualisation and reporting analytics for Quantitative Investment Strategies',
    'long_description': '# READ ME\n\n## **Analytics**\n\nTO DO\n\n\n',
    'author': 'Artur Sepp',
    'author_email': 'artursepp@gmail.com',
    'maintainer': 'Artur Sepp',
    'maintainer_email': 'artursepp@gmail.com',
    'url': 'https://github.com/ArturSepp/QuantInvestStrats',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



"""
times
"""
import time
import functools
import numpy as np


def timer(func):
    """Print the runtime of the decorated function"""
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()    # 1
        value = func(*args, **kwargs)
        end_time = time.perf_counter()      # 2
        run_time = end_time - start_time    # 3
        if run_time < 60.0:
            print(f"Finished {func.__name__!r} in {run_time:.4f} secs")
        else:
            minuts = np.floor(run_time/60.0)
            secs = run_time - 60.0*minuts
            print(f"Finished {func.__name__!r} in {minuts:.0f}m {secs:.0f}secs")
        return value
    return wrapper_timer


# built in
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Union, Tuple, Optional
from enum import Enum

# qis
import qis.plots.time_series as pts
import qis.plots.utils as put


def compute_drawdown(prices: pd.Series) -> pd.Series:

    if not isinstance(prices, pd.Series):
        print(prices)
        raise TypeError(f"in compute_max_dd: path_data must be series")

    max_dd = np.zeros_like(prices)
    last_peak = -np.inf

    for idx, price in enumerate(prices):
        if not np.isnan(price):
            if price > last_peak:
                last_peak = price
            if not np.isclose(last_peak, 0.0):
                max_dd[idx] = price / last_peak - 1.0
        else:
            max_dd[idx] = np.nan

    max_dd = pd.Series(data=max_dd, index=prices.index, name=prices.name)

    return max_dd


def compute_time_under_water(prices: pd.Series) -> Tuple[pd.Series, pd.Series]:

    if not isinstance(prices, pd.Series):
        raise TypeError(f"in compute_time_under_water: path_data must be series")

    max_dd = np.zeros_like(prices)
    dd_times = np.zeros_like(prices.index)

    last_peak = -np.inf
    last_dd_time = prices.index[0]
    dd_times[0] = last_dd_time
    for idx, (index, price) in enumerate(prices.items()):
        if not np.isnan(price):
            if price > last_peak:
                last_peak = price
                last_dd_time = index
            max_dd[idx] = price / last_peak - 1.0
            dd_times[idx] = last_dd_time
        else:
            max_dd[idx] = 0.0
            dd_times[idx] = index

    max_dd = pd.Series(data=max_dd, index=prices.index, name=prices.name)
    dd_times = (prices.index - pd.DatetimeIndex(dd_times)).days
    time_under_water = pd.Series(data=dd_times, index=prices.index, name=prices.name)

    return max_dd, time_under_water


def compute_drawdown_data(prices: Union[pd.DataFrame, pd.Series]) -> Union[pd.DataFrame, pd.Series]:
    if isinstance(prices, pd.Series):
        drawdown = compute_drawdown(prices=prices)
    elif isinstance(prices, pd.DataFrame):
        if len(prices.columns) > 1 and prices.columns.duplicated().any():
            raise ValueError(f"dublicated columns = {prices[prices.columns.duplicated()]}")
        drawdowns = []
        for asset_ in prices:
            drawdowns.append(compute_drawdown(prices=prices[asset_]))
        drawdown = pd.concat(drawdowns, axis=1)
    else:
        raise ValueError(f"unsuported type {type(prices)}")
    return drawdown


def compute_drawdown_time_data(prices: Union[pd.DataFrame, pd.Series]
                               ) -> Tuple[Union[pd.DataFrame, pd.Series], Union[pd.DataFrame, pd.Series]]:

    if isinstance(prices, pd.Series):
        drawdown, time_under_water = compute_time_under_water(prices=prices)

    else:
        drawdowns = []
        time_under_waters = []
        for asset in prices:
            drawdown, time_under_water = compute_time_under_water(prices=prices[asset])
            drawdowns.append(drawdown)
            time_under_waters.append(time_under_water)

        drawdown = pd.concat(drawdowns, axis=1)
        time_under_water = pd.concat(time_under_waters, axis=1)

    return drawdown, time_under_water


def compute_max_dd(prices: Union[pd.DataFrame, pd.Series]
                   ) -> np.ndarray:
    max_dd_data = compute_drawdown_data(prices=prices)
    max_dds = np.min(max_dd_data.to_numpy(), axis=0)
    return max_dds


def compute_avg_max(ds: pd.Series,
                    is_max: bool = True,
                    q: float = 0.1
                    ) -> (float, float, float, float):
    """
    compute dd statistics
    """
    if is_max:
        nan_data = np.where(ds.to_numpy() >= 0, ds.to_numpy(), np.nan)
    else:
        nan_data = np.where(ds.to_numpy() <= 0, ds.to_numpy(), np.nan)

    avg = np.nanmean(nan_data)
    if is_max:
        quant = np.nanquantile(nan_data, 1.0-q)
        nmax = np.nanmax(nan_data)
    else:
        quant = np.nanquantile(nan_data, q)
        nmax = np.nanmin(nan_data)

    last = ds.iloc[-1]

    return avg, quant, nmax, last


class DdLegendType(Enum):
    NONE = 1
    SIMPLE = 2
    DETAILED = 3


def plot_drawdown(prices: pd.DataFrame,
                  title: Optional[str] = None,
                  var_format: str = '{:.0%}',
                  dd_legend_type: DdLegendType = DdLegendType.DETAILED,
                  legend_loc: str = 'lower left',
                  y_limits: Tuple[Optional[float], Optional[float]] = (None, 0.0),
                  ax: plt.Subplot = None,
                  **kwargs
                  ) -> plt.Figure:

    max_dd_data = compute_drawdown_data(prices=prices)

    if dd_legend_type == DdLegendType.NONE:
        legend_loc = None
        legend_labels = None
    else:
        legend_labels = []
        for column in max_dd_data.columns:
            avg, quant, nmax, last = compute_avg_max(ds=max_dd_data[column], is_max=False)
            if dd_legend_type == DdLegendType.SIMPLE:
                legend_labels.append(f"{column}, max dd={var_format.format(nmax)}")
            elif dd_legend_type == DdLegendType.DETAILED:
                legend_labels.append(f"{column}, mean={var_format.format(avg)},"
                                     f" quantile_10%={var_format.format(quant)}, max={var_format.format(nmax)},"
                                     f" last={var_format.format(last)}")
            else:
                raise NotImplementedError(f"{dd_legend_type}")

    fig = pts.plot_time_series(df=max_dd_data,
                               var_format=var_format,
                               legend_loc=legend_loc,
                               legend_labels=legend_labels,
                               title=title,
                               y_limits=y_limits,
                               ax=ax,
                               **kwargs)
    return fig


def plot_rolling_time_under_water(prices: pd.DataFrame,
                                  title: Union[str, None] = None,
                                  var_format: str = '{:,.0f}',
                                  y_limits: Tuple[Optional[float], Optional[float]] = (0.0, None),
                                  ax: plt.Subplot = None,
                                  **kwargs
                                  ) -> plt.Figure:
    if isinstance(prices, pd.Series):
        prices = prices.to_frame()

    max_dd_data, time_under_water = compute_drawdown_time_data(prices=prices)

    legend_labels = []
    for column in time_under_water.columns:
        avg, quant, nmax, last = compute_avg_max(ds=time_under_water[column], is_max=True)
        legend_labels.append(f"{column}, mean={var_format.format(avg)}, "
                             f"quantile_10%={var_format.format(quant)}, max={var_format.format(nmax)},"
                             f" last={var_format.format(last)}")

    fig = pts.plot_time_series(df=time_under_water,
                               var_format=var_format,
                               legend_loc='upper left',
                               legend_labels=legend_labels,
                               title=title,
                               y_limits=y_limits,
                               ax=ax,
                               **kwargs)
    return fig


def plot_drawdown_lengths(price: pd.Series,
                          cut: int = 7,
                          date_format: str = '%d%b%Y',
                          title: Union[str, None] = None,
                          var_format: str = '{:.0%}',
                          legend_loc: str = 'upper center',
                          x_limits: Tuple[Optional[float], Optional[float]] = (0.0, None),
                          y_limits: Tuple[Optional[float], Optional[float]] = (None, 0.0),
                          ax: plt.Subplot = None,
                          **kwargs
                          ) -> plt.Figure:
    """
    plot drowdowns with x-being the days in dd
    """
    max_dd, time_under_water = compute_time_under_water(prices=price)

    # compute_run_data
    val0 = 0
    index0 = index00 = max_dd.index[0]
    dds_dict = {}
    for index, val in time_under_water.items():
        if val0 > 0 and val == 0:  # dd finished
            dds_dict[index] = dict(price_run=price[index00:index], dd_lenth=val0, is_current=False)
            index00 = index
        elif val0==0 and val == 0:  # ath
            index00 = index
        else:  # in dd
            pass
        val0 = val
        index0 = index
    # add last index
    if time_under_water[index0] > 0:
        dds_dict[index0] = dict(price_run=price[index00:index0], dd_lenth=time_under_water[index0], is_current=True)

    # sort by lenths
    sorted_dds_dict = dict(sorted(dds_dict.items(), key=lambda item: item[1]['dd_lenth'], reverse=True))

    dd_datas = []
    colors = put.get_n_colors(n=cut)
    linestyles = ['dotted']*cut
    for idx, (key, dd_data) in enumerate(sorted_dds_dict.items()):
        if idx < cut:
            price_run = dd_data['price_run']
            perf = price_run/price_run[0] - 1.0
            name = f"{price_run.index[0].strftime(date_format)}-{price_run.index[-1].strftime(date_format)}," \
                   f" days={dd_data['dd_lenth']}, max={np.nanmin(perf):.0%}"
            if dd_data['is_current']:
                name = f"{name}-ongoing"
                colors[idx] = 'black'
                linestyles[idx] = 'solid'
            dd_datas.append(pd.Series(perf.to_numpy(), name=name))
        else:
            break
    dd_datas = pd.concat(dd_datas, axis=1)
    fig = pts.plot_time_series(df=dd_datas,
                               var_format=var_format,
                               legend_loc=legend_loc,
                               linestyles=linestyles,
                               legend_line_type=pts.LegendLineType.NONE,
                               x_limits=x_limits,
                               y_limits=y_limits,
                               xlabel='Days in drawdown',
                               ylabel='% performance from the last peak',
                               title=title,
                               colors=colors,
                               ax=ax,
                               **kwargs)
    return fig


class UnitTests(Enum):
    DD_DATA = 1
    DRAWDOWN_TS = 2
    ROLLING_TIME = 3
    PLOT_DD_LENTHS = 4


def run_unit_test(unit_test: UnitTests):

    from qis.data.yf_data import load_etf_data
    prices = load_etf_data().dropna()

    if unit_test == UnitTests.DD_DATA:
        prices = prices.iloc[:, 0]
        max_dd, time_under_water = compute_time_under_water(prices=prices)
        print(max_dd)
        print(time_under_water)

    elif unit_test == UnitTests.DRAWDOWN_TS:
        plot_drawdown(prices=prices)

    elif unit_test == UnitTests.ROLLING_TIME:
        plot_rolling_time_under_water(prices=prices)

    elif unit_test == UnitTests.PLOT_DD_LENTHS:
        plot_drawdown_lengths(price=prices.iloc[:, 0])

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.DRAWDOWN_TS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)
